"""
Kypher parser (derived from ruruki.parsers.cypher_parser.py).
https://s3.amazonaws.com/artifacts.opencypher.org/cypher.ebnf
"""

import sys
import os
import os.path
import pprint
import parsley
import ometa.grammar
from   kgtk.kypher.grammar import KYPHER_GRAMMAR

pp = pprint.PrettyPrinter(indent=4)


# TO DO:
# - flesh out the TO DOs below
# - add missing pattern elements
# + expression simplification and normalization (e.g., with <-x-> patterns)
# + add implied clauses (e.g., from properties)
#   - handled by query translator instead
# - handle match-pattern OR-expansion (from undirected arrows and multiple relation labels)
#   - possibly also handle this by query translator (if at all)
# - handle restrictions and unsupported elements more consistently and pervasively


GRAMMAR_FILE = os.path.join(os.path.dirname(sys.modules['kgtk.kypher.grammar'].__file__), 'grammar.py')
COMPILED_GRAMMAR_FILE = os.path.join(os.path.dirname(GRAMMAR_FILE), 'grammar_compiled.py')

def compile_grammar():
    # if we want to use this for production, we can save the grammar as a Python file like this:
    cygram = ometa.grammar.OMeta.makeGrammar(KYPHER_GRAMMAR)
    with open(COMPILED_GRAMMAR_FILE, 'w') as out:
        out.write(cygram.__loader__.source)

def load_grammar(compile=True):
    if compile:
        # re/compile if necessary:
        if not os.path.exists(COMPILED_GRAMMAR_FILE) or \
           os.path.getmtime(GRAMMAR_FILE) > os.path.getmtime(COMPILED_GRAMMAR_FILE):
            compile_grammar()
        # load it like this - there doesn't seem to be a better user API available:
        import kgtk.kypher.grammar_compiled as cgc
        cgram = cgc.createParserClass(ometa.grammar.OMetaBase, {})
        return parsley.wrapGrammar(cgram)
    else:
        # rebuild the grammar from scratch:
        return parsley.makeGrammar(KYPHER_GRAMMAR, {})

Parser = load_grammar()


"""
# The grammar has 75 output patterns:
# - there are two with varying signatures - might be a mistake
# - we are currently handling about 40
# - we should redo this from scratch (see grammar notes) - maybe later

["All", fex] ["Any", fex] ["Case", ex, cas, el] ["Create", p]
["Delete", [head] + tail] ["Expression", a, opts] ["Expression2", a, c]
["Expression3", ex1, c] ["Extract", fex, ex] ["Filter", fex]
["FilterExpression", i, w] ["IdInColl", v, ex] ["Limit", ex] ["List", ex]
["ListComprehension", fex, ex] ["Literal", l] ["Match", p, w]
["Merge", [head] + tail] ["MergeActionCreate", s] ["MergeActionMatch", s]
["NodeLabel", n] ["NodePattern", s, nl, p] ["None", fex] ["Order", [head] + tail]
["Parameter", p] ["PatternElement", np, pec] ["PatternElementChain", rp, np]
["PatternPart", v, ap] ["PropertyLookup", prop_name]
["RelationshipDetail", v, q, rt, rl, p] ["RelationshipTypes", head]

["RelationshipsPattern", la, rd, ra]  <<--
["RelationshipsPattern", np, pec]     <<--

["Remove", [head] + tail] ["RemoveItemPe", p] ["RemoveItemVar", v, nl]
["Return", d, rb] ["ReturnBody", ri, o, s, l] ["ReturnItem", ex, None]
["ReturnItem", ex, s] ["ReturnItems", [head] + tail] ["Set", [head] + tail]
["SetItem", v, ex] ["SetItemPropertyExpression", pex, ex] ["Single", fex]

["SingleQuery", [head] + tail]        <<--
["SingleQuery", m, w, r]              <<--

["Skip", ex] ["Union", sq, rq] ["UnionAll", sq, rq] ["Unwind", ex, v]
["Variable", s] ["Where", ex] ["With", d, rb, w] ["add", ex1, ex2]
["and", ex1, ex2] ["call", func, distinct, args] ["count *"] 
["div", ex1, ex2] ["eq", ex1, ex2] ["gt", ex1, ex2] ["gte", ex1, ex2]
["hat", ex1, ex2] ["is_not_null"] ["is_null"] ["lt", ex1, ex2]
["lte", ex1, ex2] ["minus", ex] ["mod", ex1, ex2] ["multi", ex1, ex2]
["neq", ex1, ex2] ["not", ex] ["or", ex1, ex2] ["slice", start, end]
["sort", ex, dir] ["sub", ex1, ex2] ["xor", ex1, ex2]
"""


### Parse tree utilities

def object_to_tree(obj):
    if hasattr(obj, 'to_tree'):
        return obj.to_tree()
    elif isinstance(obj, list):
        return [object_to_tree(elt) for elt in obj]
    elif isinstance(obj, dict):
        table = {}
        for key, val in obj.items():
            if key.startswith('_') or val is None:
                continue
            table[key] = object_to_tree(val)
        return table
    else:
        return obj

def simplify_object(obj):
    if hasattr(obj, 'simplify'):
        return obj.simplify()
    elif isinstance(obj, list):
        for i in range(len(obj)):
            obj[i] = simplify_object(obj[i])
        return obj
    elif isinstance(obj, dict):
        table = {}
        for key, val in obj.items():
            if key.startswith('_') or val is None:
                continue
            table[key] = simplify_object(val)
        for key, val in table.items():
            obj[key] = val
        return obj
    else:
        return obj

def has_element(obj, test):
    """Return True if any of the `QueryElement's in `obj'
    or any of their recursive subelements satisfy `test'.
    """
    if isinstance(obj, QueryElement):
        return test(obj) or has_element(obj.__dict__, test)
    elif isinstance(obj, list):
        for elt in obj:
            if has_element(elt, test):
                return True
    elif isinstance(obj, dict):
        for elt in obj.values():
            if has_element(elt, test):
                return True
    return False


### Object representation for Kypher ASTs (Abstract Syntax Trees)

class QueryElement(object):
    ast_name = None

    def to_tree(self):
        return (self.__class__.__name__, object_to_tree(self.__dict__))

    def simplify(self):
        simplify_object(self.__dict__)
        return self


# Atoms:

class Literal(QueryElement):
    ast_name = 'Literal'
    
    def __init__(self, query, value):
        self._query = query
        self.value = value
    
class Variable(QueryElement):
    ast_name = 'Variable'

    def __init__(self, query, name):
        self._query = query
        self.name = name
        query.variables[name] = self

class AnonymousVariable(Variable):
    pass

class Parameter(QueryElement):
    ast_name = 'Parameter'
    
    def __init__(self, query, name):
        self._query = query
        self.name = name


# Arithmetic:

class Add(QueryElement):
    ast_name = 'add'

    def __init__(self, query, arg1, arg2):
        self._query = query
        self.arg1 = intern_ast(query, arg1)
        self.arg2 = intern_ast(query, arg2)

class Sub(QueryElement):
    ast_name = 'sub'

    def __init__(self, query, arg1, arg2):
        self._query = query
        self.arg1 = intern_ast(query, arg1)
        self.arg2 = intern_ast(query, arg2)

class Minus(QueryElement):
    ast_name = 'minus'

    def __init__(self, query, arg):
        self._query = query
        self.arg = intern_ast(query, arg)

class Multi(QueryElement):
    ast_name = 'multi'

    def __init__(self, query, arg1, arg2):
        self._query = query
        self.arg1 = intern_ast(query, arg1)
        self.arg2 = intern_ast(query, arg2)

class Div(QueryElement):
    ast_name = 'div'

    def __init__(self, query, arg1, arg2):
        self._query = query
        self.arg1 = intern_ast(query, arg1)
        self.arg2 = intern_ast(query, arg2)

class Mod(QueryElement):
    ast_name = 'mod'

    def __init__(self, query, arg1, arg2):
        self._query = query
        self.arg1 = intern_ast(query, arg1)
        self.arg2 = intern_ast(query, arg2)

class Hat(QueryElement):
    ast_name = 'hat'

    def __init__(self, query, arg1, arg2):
        self._query = query
        self.arg1 = intern_ast(query, arg1)
        self.arg2 = intern_ast(query, arg2)


# Comparison:

class Eq(QueryElement):
    ast_name = 'eq'

    def __init__(self, query, arg1, arg2):
        self._query = query
        self.arg1 = intern_ast(query, arg1)
        self.arg2 = intern_ast(query, arg2)

class Neq(QueryElement):
    ast_name = 'neq'

    def __init__(self, query, arg1, arg2):
        self._query = query
        self.arg1 = intern_ast(query, arg1)
        self.arg2 = intern_ast(query, arg2)

class Lt(QueryElement):
    ast_name = 'lt'

    def __init__(self, query, arg1, arg2):
        self._query = query
        self.arg1 = intern_ast(query, arg1)
        self.arg2 = intern_ast(query, arg2)

class Gt(QueryElement):
    ast_name = 'gt'

    def __init__(self, query, arg1, arg2):
        self._query = query
        self.arg1 = intern_ast(query, arg1)
        self.arg2 = intern_ast(query, arg2)

class Lte(QueryElement):
    ast_name = 'lte'

    def __init__(self, query, arg1, arg2):
        self._query = query
        self.arg1 = intern_ast(query, arg1)
        self.arg2 = intern_ast(query, arg2)

class Gte(QueryElement):
    ast_name = 'gte'

    def __init__(self, query, arg1, arg2):
        self._query = query
        self.arg1 = intern_ast(query, arg1)
        self.arg2 = intern_ast(query, arg2)


# Boolean operators and conditionals:

class Not(QueryElement):
    ast_name = 'not'

    def __init__(self, query, arg):
        self._query = query
        self.arg = intern_ast(query, arg)

class And(QueryElement):
    ast_name = 'and'

    def __init__(self, query, arg1, arg2):
        self._query = query
        self.arg1 = intern_ast(query, arg1)
        self.arg2 = intern_ast(query, arg2)

class Or(QueryElement):
    ast_name = 'or'

    def __init__(self, query, arg1, arg2):
        self._query = query
        self.arg1 = intern_ast(query, arg1)
        self.arg2 = intern_ast(query, arg2)

class Xor(QueryElement):
    ast_name = 'xor'

    def __init__(self, query, arg1, arg2):
        self._query = query
        self.arg1 = intern_ast(query, arg1)
        self.arg2 = intern_ast(query, arg2)

class Case(QueryElement):
    ast_name = 'Case'

    def __init__(self, query, test, conditions, otherwise):
        self._query = query
        self.test = intern_ast(query, test)
        self.conditions = [(intern_ast(query, when), intern_ast(query, then_)) for when, then_ in conditions]
        self.otherwise = intern_ast(query, otherwise)


# Expressions:

class Expression(QueryElement):
    ast_name = 'Expression'

    def __init__(self, query, arg, options):
        self._query = query
        self.arg = intern_ast(query, arg)
        self.options = intern_ast(query, options)

class Expression2(QueryElement):
    ast_name = 'Expression2'

    def __init__(self, query, arg1, arg2):
        self._query = query
        self.arg1 = intern_ast(query, arg1)
        self.arg2 = intern_ast(query, arg2)

class Expression3(QueryElement):
    ast_name = 'Expression3'

    def __init__(self, query, arg1, op_and_arg2):
        self._query = query
        self.arg1 = intern_ast(query, arg1)
        op_and_arg2 = op_and_arg2[0]
        self.operator = op_and_arg2[0]
        self.arg2 = len(op_and_arg2) >= 2 and intern_ast(query, op_and_arg2[1]) or None

class PropertyLookup(QueryElement):
    ast_name = 'PropertyLookup'

    def __init__(self, query, property):
        self._query = query
        self.property = property

class Call(QueryElement):
    ast_name = 'call'

    def __init__(self, query, func, distinct, args):
        self._query = query
        self.function = func
        self.distinct = distinct is not None
        self.args = intern_ast_list(query, args)


# Lists:

class List(QueryElement):
    ast_name = 'List'

    def __init__(self, query, elements):
        self._query = query
        self.elements = intern_ast_list(query, elements)


# Patterns:

class NodeLabel(QueryElement):
    ast_name = 'NodeLabel'

    def __init__(self, query, name):
        self._query = query
        self.name = name

class NodePattern(QueryElement):
    ast_name = 'NodePattern'

    def __init__(self, query, name, labels, properties):
        self._query = query
        self.variable = name
        self.labels = intern_ast_list(query, labels)
        self.graph = None
        self.properties = None
        if properties is not None:
            if properties[0] == 'Literal':
                self.properties = {key:intern_ast(query, val) for key, val in properties[1].items()}
            else:
                # parameter:
                self.properties = intern_ast(query, properties)

    def simplify(self):
        self.variable = Variable(self._query, self.variable)
        if self.labels is not None:
            self.labels = [lab.name for lab in self.labels]
        return self

    def is_anonymous(self):
        return self.variable is None or self.variable.name is None

    def normalize_term(self, implied_clauses):
        # TO DO: handle implied clauses from properties
        query = self._query
        labels = self.labels
        assert labels is None or len(labels) == 1, 'Multiple node labels are not allowed'
        if self.is_anonymous():
            self.variable = query.create_anonymous_variable()
        return self

class RelationshipPattern(QueryElement):
    ast_name = 'RelationshipsPattern' # note the extra 's'

    def __init__(self, query, *args):
        if len(args) == 2:
            # KLUDGE: `RelationshipsPattern' is one of two outputs that get invoked with incongruent signatures:
            # - normal case: ["RelationshipsPattern", la, rd, ra]
            # - WHERE case:  ["RelationshipsPattern", np, pec]
            # We handle this here with a "change class" to `PatternElement' which is what it should be.
            # TO DO: clean this up, but it will do for now:
            #        make this return a PathPattern to make normalization work similar to MATCH
            head = args[0]
            tail = args[1] is not None and list(args[1:]) or []
            PatternElement.__init__(self, query, head, tail)
            self.__class__ = PatternElement
            return
        left_arrow, detail, right_arrow = args
        self._query = query
        self.left_arrow = left_arrow
        self.detail = intern_ast(query, detail)
        self.labels = None
        self.right_arrow = right_arrow
        # a bidirectional arrow is legal in the grammar but not legal Cypher;
        if left_arrow and right_arrow:
            raise Exception('Illegal bidirectional arrow: %s' % str(self.simplify().to_tree()))

    def simplify(self):
        self.detail = simplify_object(self.detail)
        self.variable = self.detail and self.detail.variable or None
        self.qualifier = self.detail and self.detail.qualifier or None
        # multiple labels amount to an OR in KGTK which we are addressing later:
        self.labels = self.detail and self.detail.labels or None
        self.range_ = self.detail and self.detail.range_ or None
        self.properties = self.detail and self.detail.properties or None
        # a non-directional relation amounts to an OR in KGTK which we are addressing later:
        self.arrow = (self.left_arrow or '') + '--' + (self.right_arrow or '')
        delattr(self, 'left_arrow')
        delattr(self, 'right_arrow')
        delattr(self, 'detail')
        return self

    def is_anonymous(self):
        return self.variable is None or self.variable.name is None
    
    def normalize_term(self, implied_clauses):
        # TO DO: handle implied clauses from properties
        query = self._query
        labels = self.labels
        arrow = self.arrow
        assert labels is None or len(labels) == 1, 'Multiple relationship labels are not (yet) allowed'
        assert arrow != '--', 'Undirected relationships are not (yet) allowed'
        if self.is_anonymous():
            self.variable = query.create_anonymous_variable()
        return self

class RelationshipDetail(QueryElement):
    ast_name = 'RelationshipDetail'

    def __init__(self, query, variable, qualifier, types, range_, properties):
        self._query = query
        self.variable = intern_ast(query, variable)
        self.qualifier = qualifier
        self.labels = intern_ast(query, types)
        self.range_ = range_
        self.properties = None
        if properties is not None:
            if properties[0] == 'Literal':
                self.properties = {key:intern_ast(query, val) for key, val in properties[1].items()}
            else:
                # parameter:
                self.properties = intern_ast(query, properties)

class RelationshipTypes(QueryElement):
    ast_name = 'RelationshipTypes'

    def __init__(self, query, *types):
        self._query = query
        self.labels = types

    def simplify(self):
        return self.labels

class PatternElementChain(QueryElement):
    ast_name = 'PatternElementChain'

    def __init__(self, query, head, tail):
        self._query = query
        self.head = intern_ast(query, head)
        self.tail = intern_ast(query, tail)

    def simplify(self):
        return [self.head.simplify(), self.tail.simplify()]
        
class PatternElement(QueryElement):
    ast_name = 'PatternElement'

    def __init__(self, query, element, chain):
        self._query = query
        self.head = intern_ast(query, element)
        self.tail = intern_ast_list(query, chain)
        self.graph = None

    def simplify(self):
        chain = [self.head.simplify()]
        chain[0].graph = self.graph
        for subchain in simplify_object(self.tail):
            chain += subchain
        return chain
    
class PatternPart(QueryElement):
    ast_name = 'PatternPart'

    def __init__(self, query, variable, pattern):
        self._query = query
        # these can either be named with a variable or anonymous:
        self.variable = intern_ast(query, variable)
        self.pattern = intern_ast(query, pattern)

    def simplify(self):
        # simply rename to `PathPattern':
        return PathPattern(self._query, simplify_object(self.variable), simplify_object(self.pattern))

class GraphPatternPart(QueryElement):
    ast_name = 'GraphPatternPart'

    def __init__(self, query, graph, pattern):
        self._query = query
        # we have a pattern prefixed with a graph variable:
        self.graph = intern_ast(query, graph)
        self.pattern = intern_ast(query, pattern)

    def simplify(self):
        # simply rename to `PathPattern' and set graph variable:
        pattern = PathPattern(self._query, None, simplify_object(self.pattern))
        pattern.graph = simplify_object(self.graph)
        return pattern

class GraphRelationshipsPattern(PatternElement):
    ast_name = 'GraphRelationshipsPattern'

    def __init__(self, query, graph, head, tail):
        # this is a variant of RelationshipPattern where we don't have the ambiguity:
        super().__init__(query, head, [tail])
        self.graph = intern_ast(query, graph)

class PathPattern(QueryElement):

    def __init__(self, query, variable, pattern):
        self._query = query
        # these can either be named with a variable or anonymous:
        self.variable = variable
        self.pattern = pattern
        self.graph = None

    def normalize_clauses(self):
        # Path patterns are either N, N-R-N, or N-R-N-R-...-R-N for two or more relations.
        # We normalize them onto a single or list of implicitly conjoined N-R-N path patterns.
        # Assumes simplification has been run.
        query = self._query
        pattern = self.pattern
        graph = self.graph
        num_elements = len(pattern)
        if num_elements == 3:
            if pattern[1].arrow == '<--':
                # normalize onto forward direction:
                head = pattern[0]
                pattern[0] = pattern[2]
                pattern[2] = head
                pattern[1].arrow = '-->'
            pattern[0].graph = graph
            return self
        elif num_elements == 1:
            pattern[0].graph = graph
            relpat = RelationshipPattern(query, None, None, None)
            relpat.arrow = '-->'
            relpat.variable = query.create_anonymous_variable()
            self.pattern.append(relpat)
            nodepat = NodePattern(query, None, None, None)
            nodepat.variable = query.create_anonymous_variable()
            self.pattern.append(nodepat)
            return self
        else:
            assert num_elements >= 5 and num_elements % 2 == 1, 'Unexpected number of path pattern elements'
            norm_patterns = []
            while len(pattern) >= 3:
                subpat = PathPattern(query, self.variable, pattern[0:3])
                subpat.graph = graph
                norm_patterns.append(subpat.normalize_clauses())
                # we create a connecting variable node, but we do not copy any of the other attributes if any,
                # since those might result in additional clauses which we only want to create once:
                conn_nodepat = pattern[2]
                if conn_nodepat.is_anonymous():
                    conn_nodepat.variable = query.create_anonymous_variable()
                nodepat = NodePattern(query, None, None, None)
                nodepat.variable = conn_nodepat.variable
                nodepat.graph = graph
                conn_nodepat = nodepat
                pattern = [conn_nodepat] + pattern[3:]
            return norm_patterns

    def normalize_terms(self, implied_clauses):
        query = self._query
        pattern = self.pattern
        assert len(pattern) == 3, 'Unnormalized path pattern'
        pattern[0] = pattern[0].normalize_term(implied_clauses)
        pattern[1] = pattern[1].normalize_term(implied_clauses)
        pattern[2] = pattern[2].normalize_term(implied_clauses)

    def normalize(self):
        # this gives us either a single or list of normalized N-R-N PathPattern's:
        clauses = self.normalize_clauses()
        if not hasattr(clauses, '__iter__'):
            clauses = [clauses]
        norm_clauses = []
        for clause in clauses:
            norm_clauses.append(clause)
            clause.normalize_terms(norm_clauses)
        return norm_clauses


# Query top level clauses:

class Match(QueryElement):
    ast_name = 'Match'

    def __init__(self, query, pattern, where):
        self._query = query
        self.pattern = intern_ast_list(query, pattern)
        self.where = intern_ast(query, where)

class Where(QueryElement):
    ast_name = 'Where'

    def __init__(self, query, expression):
        self._query = query
        self.expression = intern_ast(query, expression)


class Skip(QueryElement):
    ast_name = 'Skip'

    def __init__(self, query, expression):
        self._query = query
        self.expression = intern_ast(query, expression)

class Limit(QueryElement):
    ast_name = 'Limit'

    def __init__(self, query, expression):
        self._query = query
        self.expression = intern_ast(query, expression)

class Order(QueryElement):
    ast_name = 'Order'

    def __init__(self, query, items):
        self._query = query
        self.items = intern_ast(query, items)

class SortItem(QueryElement):
    ast_name = 'sort'

    def __init__(self, query, expression, direction):
        self._query = query
        self.expression = intern_ast(query, expression)
        self.direction = direction

class ReturnItem(QueryElement):
    ast_name = 'ReturnItem'

    def __init__(self, query, expression, name):
        self._query = query
        self.name = name
        self.expression = intern_ast(query, expression)

class ReturnItems(QueryElement):
    ast_name = 'ReturnItems'

    def __init__(self, query, items):
        self._query = query
        self.items = []
        if len(items) > 0 and items[0] == '*':
            self.items.append(ReturnItem(query, Variable(query, '*'), None))
            items = items[1:]
        if len(items) > 0:
            items = intern_ast(query, items)
        self.items += items

class ReturnBody(QueryElement):
    ast_name = 'ReturnBody'

    def __init__(self, query, items, order, skip, limit):
        self._query = query
        self.items = intern_ast(query, items)
        self.order = intern_ast(query, order)
        self.skip = intern_ast(query, skip)
        self.limit = intern_ast(query, limit)

    def simplify(self):
        self.items = self.items.items
        return self

class Return(QueryElement):
    ast_name = 'Return'

    def __init__(self, query, distinct, body):
        self._query = query
        self.distinct = distinct is not None
        self.body = intern_ast(query, body)

    def simplify(self):
        body = self.body.simplify()
        self.items = body.items
        self.order = body.order
        self.skip = body.skip
        self.limit = body.limit
        delattr(self, 'body')
        return self


class SingleQuery(QueryElement):
    ast_name = 'SingleQuery'

    def __init__(self, query, match, with_, return_):
        """TO DO: COMPLETE"""
        self._query = query
        self.match = intern_ast(query, match)
        self.return_ = intern_ast(query, return_)


### AST internment:

def build_ast_name_table():
    table = {}
    module = sys.modules[SingleQuery.__module__]
    for elt in dir(module):
        elt = getattr(module, elt)
        if hasattr(elt, 'ast_name'):
            table[getattr(elt, 'ast_name')] = elt
    return table

AST_NAME_TABLE = build_ast_name_table()

def intern_ast(query, ast):
    if ast is None:
        return None
    elif isinstance(ast, QueryElement):
        return ast
    elif isinstance(ast, list) and len(ast) > 0:
        if isinstance(ast[0], list):
            return [intern_ast(query, elt) for elt in ast]
        else:
            klass = AST_NAME_TABLE.get(ast[0])
            if klass is not None:
                return klass(query, *ast[1:])
    raise Exception('Unhandled expression type: %s' % ast)

def intern_ast_list(query, ast_list):
    if ast_list is None:
        return None
    elif isinstance(ast_list, list):
        return [intern_ast(query, ast) for ast in ast_list]
    raise Exception('Unhandled list type: %s' % ast_list)


### Kypher query:

class KypherQuery(object):
    def __init__(self, query_string):
        self.query = None
        self.variables = {}
        self.simplified = False
        self.match_clauses = None
        self.parse = Parser(query_string)
        self.query = intern_ast(self, self.parse.Kypher())

    def to_tree(self):
        return (self.__class__.__name__, self.query and self.query.to_tree() or None)

    def simplify(self):
        """Simplifies the structure of the interned parse tree to get rid of unneeded
        elements and nesting that are simply an artifact of the grammar specification.
        """
        if not self.simplified:
            self.query = self.query.simplify()
            self.simplified = True
        return self

    def get_match_pattern(self):
        assert isinstance(self.query, SingleQuery), 'Only single-match queries are supported, no unions'
        return self.query.match.pattern

    def get_match_clauses(self):
        """Returns a list of [<NodePattern ...> <RelationshipPattern --> ...> <NodePattern ...>] clauses
        where each pattern element has a named or anonymous variable and optional single label.  All
        property elements have been translated into additional normalized match clauses (not yet).
        """
        if self.match_clauses is None:
            self.simplify()
            self.match_clauses = []
            current_graph = None
            for pathpat in self.get_match_pattern():
                for normpath in pathpat.normalize():
                    normpath = normpath.pattern
                    current_graph = normpath[0].graph or current_graph
                    normpath[0].graph = current_graph
                    self.match_clauses.append(normpath)
        return self.match_clauses

    def get_where_clause(self):
        assert isinstance(self.query, SingleQuery), 'Only single-match queries are supported, no unions'
        self.simplify()
        where = self.query.match.where
        return where

    def get_return_clause(self):
        assert isinstance(self.query, SingleQuery), 'Only single-match queries are supported, no unions'
        self.simplify()
        ret = self.query.return_
        return ret

    def get_order_clause(self):
        ret = self.get_return_clause()
        order = ret and ret.order or None
        return order

    def get_skip_clause(self):
        ret = self.get_return_clause()
        skip = ret and ret.skip or None
        return skip

    def get_limit_clause(self):
        ret = self.get_return_clause()
        limit = ret and ret.limit or None
        return limit

    def create_anonymous_variable(self):
        i = 1
        while True:
            varname = '_x%04d' % i
            if varname not in self.variables:
                var = AnonymousVariable(self, varname)
                self.variables[varname] = var
                return var
            i += 1


### Top level:

def parse(query_string):
    return Parser(query_string).Kypher()

def intern(query_string):
    return KypherQuery(query_string)


# Example w/ interesting differences between node and relation patterns:
# - node patterns can have multiple labels (and-ed), all with a prefixing colon
# - relation patterns can have alternative labels, only the first prefixed with a colon
#
# cq = cp.intern("MATCH (n:Person :Human {name: 'Bob'})-[:Loves | Adores {id: '17'}]->(n2) RETURN DISTINCT n;")



import shutil
import unittest
import tempfile
import shlex
from   io import StringIO
import pandas as pd
from kgtk.cli_entry import cli_entry


# TO DO:
# - file aliasing via --as, realiasing
# - files not required to exist after initial import
# - queries from stdin (query pipes not yet supported)
# - NULL value tests and conversion functions

class TestKGTKQuery(unittest.TestCase):
    def setUp(self) -> None:
        self.file_path = 'data/kypher/graph.tsv'
        self.file_path_gz = 'data/kypher/graph.tsv.gz'
        self.file_path_bz2 = 'data/kypher/graph.tsv.bz2'
        self.quals_path = 'data/kypher/quals.tsv'
        self.works_path = 'data/kypher/works.tsv'
        self.props_path = 'data/kypher/props.tsv'
        self.literals_path = 'data/kypher/literals.tsv'
        self.temp_dir = tempfile.mkdtemp()
        self.sqldb = f'{self.temp_dir}/test.sqlite3.db'
        self.df = pd.read_csv(self.file_path, sep='\t')

    def tearDown(self) -> None:
        shutil.rmtree(self.temp_dir)

    def test_kgtk_query_default(self):
        cli_entry("kgtk", "query", "-i", self.file_path, "-o", f'{self.temp_dir}/out.tsv', '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 9)

    def test_kgtk_query_default_gzip(self):
        cli_entry("kgtk", "query", "-i", self.file_path_gz, "-o", f'{self.temp_dir}/out.tsv.gz', '--graph-cache',
                  self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv.gz', sep='\t')
        self.assertTrue(len(df) == 9)

    def test_kgtk_query_default_bz2(self):
        cli_entry("kgtk", "query", "-i", self.file_path_bz2, "-o", f'{self.temp_dir}/out.tsv.bz2', '--graph-cache',
                  self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv.bz2', sep='\t')
        self.assertTrue(len(df) == 9)

    def test_kgtk_query_match(self):
        cli_entry("kgtk", "query", "-i", self.file_path, "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "(i)-[:loves]->(c)", '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 3)
        ids = list(df['id'].unique())
        self.assertTrue('e11' in ids)
        self.assertTrue('e12' in ids)
        self.assertTrue('e14' in ids)

    def test_kgtk_query_limit(self):
        cli_entry("kgtk", "query", "-i", self.file_path, "-o", f'{self.temp_dir}/out.tsv', "--limit",
                  "3", '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 3)
        ids = list(df['id'].unique())
        self.assertTrue('e11' in ids)
        self.assertTrue('e12' in ids)
        self.assertTrue('e13' in ids)

    def test_kgtk_query_limit_skip(self):
        cli_entry("kgtk", "query", "-i", self.file_path, "-o", f'{self.temp_dir}/out.tsv', "--limit",
                  "3", "--skip", "2", '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 3)
        ids = list(df['id'].unique())
        self.assertTrue('e13' in ids)
        self.assertTrue('e14' in ids)
        self.assertTrue('e21' in ids)

    def test_kgtk_query_hans_filter(self):
        cli_entry("kgtk", "query", "-i", self.file_path, "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "(:Hans)-[]->()", '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 2)
        ids = list(df['id'].unique())
        self.assertTrue('e11' in ids)
        self.assertTrue('e21' in ids)

    def test_kgtk_query_otto_name_filter(self):
        cli_entry("kgtk", "query", "-i", self.file_path, "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "(:Otto)-[:name]->()", '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 1)
        ids = list(df['id'].unique())
        self.assertTrue('e22' in ids)

    def test_kgtk_query_where_double_letter(self):
        cli_entry("kgtk", "query", "-i", self.file_path, "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "(p)-[:name]->(n)", "--where", 'n =~".*(.)\\\\1.*"', '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 2)
        ids = list(df['id'].unique())
        self.assertTrue('e22' in ids)
        self.assertTrue('e24' in ids)

    def test_kgtk_query_where_IN(self):
        cli_entry("kgtk", "query", "-i", self.file_path, "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "(p)-[:name]->(n)", "--where", 'p IN ["Hans", "Susi"]', '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 2)
        ids = list(df['id'].unique())
        self.assertTrue('e21' in ids)
        self.assertTrue('e25' in ids)

    def test_kgtk_query_where_upper_substring(self):
        cli_entry("kgtk", "query", "-i", self.file_path, "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "(p)-[:name]->(n)", "--where", "upper(substr(n,2,1)) >= 'J'", '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 4)
        ids = list(df['id'].unique())
        self.assertTrue('e22' in ids)
        self.assertTrue('e23' in ids)
        self.assertTrue('e24' in ids)
        self.assertTrue('e25' in ids)

    def test_kgtk_query_where_upper_substring_sorted(self):
        cli_entry("kgtk", "query", "-i", self.file_path, "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "(p)-[:name]->(n)", "--where", "upper(substr(n,2,1)) >= 'J'", "--order-by", "substr(n,2,1)",
                  '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 4)
        ids = list(df['id'].unique())
        self.assertTrue('e23' == ids[0])
        self.assertTrue('e24' == ids[1])
        self.assertTrue('e22' == ids[2])
        self.assertTrue('e25' == ids[3])

    def test_kgtk_query_where_upper_substring_sorted_desc(self):
        cli_entry("kgtk", "query", "-i", self.file_path, "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "(p)-[:name]->(n)", "--where", "upper(substr(n,2,1)) >= 'J'", "--order-by", "substr(n,2,1) desc",
                  '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 4)
        ids = list(df['id'].unique())
        self.assertTrue('e25' == ids[0])
        self.assertTrue('e22' == ids[1])
        self.assertTrue('e24' == ids[2])
        self.assertTrue('e23' == ids[3])

    def test_kgtk_query_select_columns(self):
        cli_entry("kgtk", "query", "-i", self.file_path, "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "(p)-[:name]->(n)", "--where", "upper(substr(n,2,1)) >= 'J'", "--return", "p,n", '--graph-cache',
                  self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 4)
        columns = list(df.columns)
        self.assertTrue('node1' in columns)
        self.assertTrue('node2' in columns)
        node1s = list(df['node1'].unique())
        self.assertTrue('Otto' in node1s)
        self.assertTrue('Joe' in node1s)
        self.assertTrue('Molly' in node1s)
        self.assertTrue('Susi' in node1s)

    def test_kgtk_query_switch_columns(self):
        cli_entry("kgtk", "query", "-i", self.file_path, "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "(p)-[r:name]->(n)", "--where", "upper(substr(n,2,1)) >= 'J'", "--return", "p,n, r, r.label",
                  '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 4)
        columns = list(df.columns)
        self.assertTrue('node1' in columns)
        self.assertTrue('node2' in columns)
        self.assertTrue('id' in columns)
        self.assertTrue('label' in columns)
        node1s = list(df['node1'].unique())
        self.assertTrue('Otto' in node1s)
        self.assertTrue('Joe' in node1s)
        self.assertTrue('Molly' in node1s)
        self.assertTrue('Susi' in node1s)

    def test_kgtk_query_return_columns_modify_functions(self):
        cli_entry("kgtk", "query", "-i", self.file_path, "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "(p)-[r:name]->(n)", "--where", "upper(substr(n,2,1)) >= 'J'",
                  "--return", "lower(p) as node1, r.label, n, r",
                  '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 4)
        columns = list(df.columns)
        self.assertTrue('node1' in columns)
        self.assertTrue('node2' in columns)
        self.assertTrue('id' in columns)
        self.assertTrue('label' in columns)
        node1s = list(df['node1'].unique())
        self.assertTrue('otto' in node1s)
        self.assertTrue('joe' in node1s)
        self.assertTrue('molly' in node1s)
        self.assertTrue('susi' in node1s)

    def test_kgtk_query_kgtk_unstringify(self):
        cli_entry("kgtk", "query", "-i", self.file_path, "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "(p)-[r:name]->(n)", "--where", "upper(substr(n,2,1)) >= 'J'",
                  "--return", "p, r.label, kgtk_unstringify(n) as node2, r",
                  '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 4)
        columns = list(df.columns)
        self.assertTrue('node2' in columns)
        self.assertTrue('node1' in columns)
        self.assertTrue('id' in columns)
        self.assertTrue('label' in columns)
        vals = list(df['node2'].unique())
        self.assertTrue('Molly' in vals)

    def test_kgtk_query_para(self):
        cli_entry("kgtk", "query", "-i", self.file_path, "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "(p)-[r:name]->(n)", "--where", " n = $name OR n = $name2 OR n = $name3 ",
                  "--para", "name=\"'Hans'@de\"", "--spara", "name2=Susi", "--lqpara", "name3=Otto@de", '--graph-cache',
                  self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 2)
        columns = list(df.columns)
        self.assertTrue('node2' in columns)
        self.assertTrue('node1' in columns)
        self.assertTrue('id' in columns)
        self.assertTrue('label' in columns)
        ids = list(df['id'].unique())
        self.assertTrue('e25' in ids)
        self.assertTrue('e22' in ids)

    def test_kgtk_query_lgstring_land(self):
        cli_entry("kgtk", "query", "-i", self.file_path, "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "(p)-[r:name]->(n)", "--where", 'n.kgtk_lqstring_lang = "de"', '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 2)
        ids = list(df['id'].unique())
        self.assertTrue('e21' in ids)
        self.assertTrue('e22' in ids)

    def test_kgtk_query_reflexive_edges(self):
        cli_entry("kgtk", "query", "-i", self.file_path, "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "(a)-[]->(a)", '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 1)
        ids = list(df['id'].unique())
        self.assertTrue('e14' in ids)

    def test_kgtk_query_multi_step_path(self):
        cli_entry("kgtk", "query", "-i", self.file_path, "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "(na)<-[:name]-(a)-[r:loves]->(b)-[:name]->(nb)", "--return", "r, na, r.label, nb", '--graph-cache',
                  self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 3)
        ids = list(df['id'].unique())
        node2s = list(df['node2'].unique())
        node2_1s = list(df['node2.1'].unique())
        self.assertTrue('e14' in ids)
        self.assertTrue('e11' in ids)
        self.assertTrue('e12' in ids)
        self.assertTrue('Joe' in node2s)
        self.assertTrue("'Hans'@de" in node2s)
        self.assertTrue("'Otto'@de" in node2s)
        self.assertTrue('Joe' in node2_1s)
        self.assertTrue('Molly' in node2_1s)
        self.assertTrue('Susi' in node2_1s)

    def test_kgtk_query_multi_step_path_german_lovers(self):
        cli_entry("kgtk", "query", "-i", self.file_path, "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "(na)<-[:name]-(a)-[r:loves]->(b)-[:name]->(nb)",
                  "--where", 'na.kgtk_lqstring_lang = "de" OR nb.kgtk_lqstring_lang = "de"',
                  "--return", "r, na, r.label, nb", '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 2)
        ids = list(df['id'].unique())
        node2s = list(df['node2'].unique())
        node2_1s = list(df['node2.1'].unique())
        self.assertTrue('e11' in ids)
        self.assertTrue('e12' in ids)
        self.assertTrue("'Hans'@de" in node2s)
        self.assertTrue("'Otto'@de" in node2s)
        self.assertTrue('Molly' in node2_1s)
        self.assertTrue('Susi' in node2_1s)

    def test_kgtk_query_multi_step_path_german_lovers_anonymous(self):
        cli_entry("kgtk", "query", "-i", self.file_path, "-o", f'{self.temp_dir}/out.tsv', "--match",
                  # test connection through anonymous node variables instead of a and b:
                  "(na)<-[:name]-()-[r:loves]->()-[:name]->(nb)",
                  "--where", 'na.kgtk_lqstring_lang = "de" OR nb.kgtk_lqstring_lang = "de"',
                  "--return", "r, na, r.label, nb", '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 2)
        ids = list(df['id'].unique())
        node2s = list(df['node2'].unique())
        node2_1s = list(df['node2.1'].unique())
        self.assertTrue('e11' in ids)
        self.assertTrue('e12' in ids)
        self.assertTrue("'Hans'@de" in node2s)
        self.assertTrue("'Otto'@de" in node2s)
        self.assertTrue('Molly' in node2_1s)
        self.assertTrue('Susi' in node2_1s)

    def test_kgtk_query__named_multi_graph_join(self):
        cli_entry("kgtk", "query", "-i", self.file_path,
                  "-i", self.works_path,
                  "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "g: (x)-[:loves]->(y), w: (y)-[:works]->(c)", '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 3)
        for i, row in df.iterrows():
            if row['id'] == 'e14':
                self.assertEqual(row['node1'], 'Joe')
                self.assertEqual(row['node2'], 'Joe')
                self.assertEqual(row['id.1'], 'w13')
                self.assertEqual(row['node1.1'], 'Joe')
                self.assertEqual(row['label.1'], 'works')
                self.assertEqual(row['node2.1'], 'Kaiser')
                self.assertEqual(row['node1;salary'], 20000)
                self.assertEqual(row['graph'], 'employ')

    def test_kgtk_query_default_multi_graph_join(self):
        cli_entry("kgtk", "query", "-i", self.file_path,
                  "-i", self.works_path,
                  "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "(x)-[:loves]->(y), w: (y)-[:works]->(c)", '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 3)
        for i, row in df.iterrows():
            if row['id'] == 'e14':
                self.assertEqual(row['node1'], 'Joe')
                self.assertEqual(row['node2'], 'Joe')
                self.assertEqual(row['id.1'], 'w13')
                self.assertEqual(row['node1.1'], 'Joe')
                self.assertEqual(row['label.1'], 'works')
                self.assertEqual(row['node2.1'], 'Kaiser')
                self.assertEqual(row['node1;salary'], 20000)
                self.assertEqual(row['graph'], 'employ')

    def test_kgtk_query_default_multi_graph_join_kgtk_compliant(self):
        cli_entry("kgtk", "query", "-i", self.file_path,
                  "-i", self.works_path,
                  "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "g: (x)-[r:loves]->(y), w: (y)-[:works]->(c)",
                  "--return", 'r, x, r.label, y as node2, c as `node2;work`', '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 3)
        for i, row in df.iterrows():
            if row['id'] == 'e11':
                self.assertEqual(row['node1'], 'Hans')
                self.assertEqual(row['node2'], 'Molly')
                self.assertEqual(row['node2;work'], 'Renal')

    def test_kgtk_query_default_multi_graph_join_property_access_restriction_cast_integer(self):
        cli_entry("kgtk", "query", "-i", self.file_path,
                  "-i", self.works_path,
                  "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "g: (x)-[r:loves]->(y), w: (y {salary: s})-[:works]->(c)",
                  "--where", "cast(s, integer) >= 10000",
                  "--return", 'r, x, r.label, y as node2, c as `node2;work`, s as `node2;salary`', '--graph-cache',
                  self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')

        self.assertTrue(len(df) == 2)
        for i, row in df.iterrows():
            if row['id'] == 'e11':
                self.assertEqual(row['node1'], 'Hans')
                self.assertEqual(row['node2'], 'Molly')
                self.assertEqual(row['node2;work'], 'Renal')

    def test_kgtk_query_multi_graph_ambiguous_return_alias(self):
        # Tests bug fix for case where 'node1'/'node2' return aliases are ambiguous for the two input graphs..
        cli_entry("kgtk", "query",
                  "-i", self.file_path, "-i", self.works_path, "-o", f'{self.temp_dir}/out.tsv',
                  "--match", "g: (:Susi)<-[:loves]-(n1), w: (n1)-[prop]->(n2)",
                  "--return", "prop, prop.label, n1 as node1, n2 as node2",
                  "--order-by", "node2",
                  "--graph-cache", self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 2)
        for i, row in df.iterrows():
            if row['id'] == 'w22':
                self.assertEqual(row['node1'], 'Otto')
                self.assertEqual(row['label'], 'department')
                self.assertEqual(row['node2'], 'Pharm')
                
    def test_kgtk_query_max(self):
        cli_entry("kgtk", "query", "-i", self.file_path,
                  "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "g: (x)-[r]->(y)",
                  "--return", 'max(x) as node1, r.label, y, r', '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 1)
        for i, row in df.iterrows():
            if row['id'] == 'e25':
                self.assertEqual(row['node1'], 'Susi')
                self.assertEqual(row['node2'], 'Susi')

    def test_kgtk_query_max_x_per_r(self):
        cli_entry("kgtk", "query", "-i", self.file_path,
                  "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "g: (x)-[r]->(y)",
                  "--return", 'r, max(x), r.label, y',
                  "--limit", "5", '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')

        self.assertTrue(len(df) == 5)
        ids = list(df['id'].unique())
        self.assertTrue('e11' in ids)
        self.assertTrue('e12' in ids)
        self.assertTrue('e13' in ids)
        self.assertTrue('e14' in ids)
        self.assertTrue('e21' in ids)

    def test_kgtk_query_count(self):
        cli_entry("kgtk", "query", "-i", self.file_path,
                  "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "g: (x)-[r]->(y)",
                  "--where", 'x = "Joe"',
                  "--return", 'count(x) as N', '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')

        self.assertTrue(len(df) == 1)
        for i, row in df.iterrows():
            self.assertEqual(row['N'], 3)

    def test_kgtk_query_count_distinct(self):
        cli_entry("kgtk", "query", "-i", self.file_path,
                  "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "g: (x)-[r]->(y)",
                  "--where", 'x = "Joe"',
                  "--return", 'count(distinct x) as N', '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 1)
        for i, row in df.iterrows():
            self.assertEqual(row['N'], 1)

    def test_kgtk_query_biggest_salary(self):
        cli_entry("kgtk", "query", "-i", self.works_path,
                  "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "w: (y {salary: s})-[r:works]->(c)",
                  "--return", 'max(cast(s, int)) as `node1;salary`, y, "works" as label, c, r', '--graph-cache',
                  self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 1)
        for i, row in df.iterrows():
            self.assertEqual(row['node1;salary'], 20000)
            self.assertEqual(row['node1'], 'Joe')
            self.assertEqual(row['node2'], 'Kaiser')

    def test_kgtk_query_date_filter(self):
        cli_entry("kgtk", "query", "-i", self.quals_path, "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "(eid)-[q]->(time)", "--where", "time.kgtk_date_year < 2005", '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 4)
        ids = list(df['id'].unique())
        self.assertTrue('m11' in ids)
        self.assertTrue('m12' in ids)
        self.assertTrue('m13' in ids)
        self.assertTrue('m14' in ids)

    def test_kgtk_query_three_graphs(self):
        cli_entry("kgtk", "query", "-i", self.works_path,
                  "-i", self.quals_path,
                  "-i", self.props_path,
                  "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "work: (x)-[r {label: rl}]->(y), qual: (r)-[rp {label: p}]->(time), prop: (p)-[:member]->(:set1)",
                  "--where", 'time.kgtk_date_year <= 2000',
                  "--return", 'r as id, x, rl, y, p as trel, time as time', '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 3)
        for i, row in df.iterrows():
            if row['id'] == 'w12':
                self.assertEqual(row['node1'], 'Otto')
                self.assertEqual(row['node2'], 'Kaiser')
                self.assertEqual(row['trel'], 'ends')
                self.assertEqual(row['time'], '^1987-11-08T04:56:34Z/10')

    def test_kgtk_query_property_enumeration_list(self):
        cli_entry("kgtk", "query", "-i", self.works_path,
                  "-i", self.quals_path,
                  "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "work: (x)-[r {label: rl}]->(y), qual: (r)-[rp {label: p}]->(time)",
                  "--where", "p in ['starts', 'ends'] and time.kgtk_date_year <= 2000",
                  "--return", 'r as id, x, rl, y, p as trel, time as time', '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 3)
        for i, row in df.iterrows():
            if row['id'] == 'w12':
                self.assertEqual(row['node1'], 'Otto')
                self.assertEqual(row['node2'], 'Kaiser')
                self.assertEqual(row['trel'], 'ends')
                self.assertEqual(row['time'], '^1987-11-08T04:56:34Z/10')

            if row['id'] == 'w11':
                self.assertEqual(row['node1'], 'Hans')
                self.assertEqual(row['node2'], 'ACME')
                self.assertEqual(row['trel'], 'starts')
                self.assertEqual(row['time'], '^1984-12-17T00:03:12Z/11')

    def test_kgtk_query_multi_graph_regex(self):
        cli_entry("kgtk", "query", "-i", self.works_path,
                  "-i", self.quals_path,
                  "-o", f'{self.temp_dir}/out.tsv', "--match",
                  "work: (x)-[r {label: rl}]->(y), qual: (r)-[rp {label: p}]->(time)",
                  "--where", "p =~ 's.*' and time.kgtk_date_year <= 2000",
                  "--return", 'r as id, x, rl, y, p as trel, time as time',
                  "--order-by", 'p desc, time asc', '--graph-cache', self.sqldb)
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 2)
        for i, row in df.iterrows():
            if row['id'] == 'w13':
                self.assertEqual(row['node1'], 'Joe')
                self.assertEqual(row['node2'], 'Kaiser')
                self.assertEqual(row['trel'], 'starts')
                self.assertEqual(row['time'], '^1996-02-23T08:02:56Z/09')

            if row['id'] == 'w11':
                self.assertEqual(row['node1'], 'Hans')
                self.assertEqual(row['node2'], 'ACME')
                self.assertEqual(row['trel'], 'starts')
                self.assertEqual(row['time'], '^1984-12-17T00:03:12Z/11')

    def test_kgtk_query_mod_operator(self):
        cli_entry('kgtk', 'query', '-i', self.file_path,
                  '--graph-cache', self.sqldb,
                  '-o', f'{self.temp_dir}/out.tsv',
                  '--match', '(n1)-[r:name]->(n2)',
                  '--return', 'r, n1, r.label, n2, length(n2) % 3 as rem')
                  
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 5)
        for i, row in df.iterrows():
            if row['id'] == 'e21':
                self.assertEqual(int(row['rem']), 0)
            if row['id'] == 'e22':
                self.assertEqual(int(row['rem']), 0)
            if row['id'] == 'e23':
                self.assertEqual(int(row['rem']), 2)
            if row['id'] == 'e24':
                self.assertEqual(int(row['rem']), 1)
            if row['id'] == 'e25':
                self.assertEqual(int(row['rem']), 0)

    def test_kgtk_query_order_by_alias(self):
        cli_entry('kgtk', 'query', '-i', self.file_path,
                  '--graph-cache', self.sqldb,
                  '-o', f'{self.temp_dir}/out.tsv',
                  '--match', '(n1)-[r:name]->(n2)',
                  '--return', 'r, n1, r.label, n2, length(n2) as `node2;len`',
                  '--order-by', '`node2;len`')
                  
        df = pd.read_csv(f'{self.temp_dir}/out.tsv', sep='\t')
        self.assertTrue(len(df) == 5)
        for i, row in df.iterrows():
            if i == 0:
                self.assertEqual(int(row['node2;len']), 5)
            if i == 1:
                self.assertEqual(int(row['node2;len']), 6)
            if i == 2:
                self.assertEqual(int(row['node2;len']), 7)
            if i == 3:
                self.assertEqual(int(row['node2;len']), 9)
            if i == 4:
                self.assertEqual(int(row['node2;len']), 9)


    ### Testing literal accessors:
    
    def run_literal_access_query(self, query, literals=None, output=None, db=None):
        """Run KGTK 'query' command on literals test data and return the result as a dataframe.
        """
        literals = literals or self.literals_path
        output = output or f'{self.temp_dir}/out.tsv'
        db = db or self.sqldb
        query = query.format(LITERALS=literals, OUTPUT=output, DB=db)
        cli_entry(*shlex.split(query))
        df = pd.read_csv(output, sep='\t')
        return df

    def result_to_dataframe(self, result):
        """Convert one or more 'result' rows into a dataframe.
        """
        if isinstance(result, str):
            result = [result]
        out = StringIO()
        for line in result:
            out.write(line)
            if not line.endswith('\n'):
                out.write('\n')
        df = pd.read_csv(StringIO(out.getvalue()), sep='\t')
        return df

    def assert_literal_access_query_result(self, query, result, rowids=None,
                                           literals=None, output=None, db=None):
        """Assert that the dataframes produced by 'query' and defined by 'result' are the same.
        If 'rowids' is not None, only compare the specified rows.  This currently assumes that
        the rows are in the same order; we could generalize that down the line if necessary.
        """
        qdf = self.run_literal_access_query(query)
        rdf = self.result_to_dataframe(result)
        if rowids is not None:
            if not isinstance(rowids, list):
                rowids = [rowids]
            qdf = qdf.loc[qdf['id'].isin(rowids)]
            rdf = rdf.loc[rdf['id'].isin(rowids)]
        self.assertTrue(qdf.equals(rdf))

    def test_kgtk_query_literal_access_kgtk_string(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:sy1)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_string(v) as `node2;is_string`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;is_string""",
                """esy1\tsy1\tsymbol\tFooBar\t0"""]
        self.assert_literal_access_query_result(query, result)

        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:st1)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_string(v) as `node2;is_string`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;is_string""",
                """est1\tst1\tstring\t"Franz Klammer"\t1"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_stringify(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:sy1)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_stringify(v) as `node2;as_string`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;as_string""",
                """esy1\tsy1\tsymbol\tFooBar\t"FooBar\""""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_unstringify(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:st2)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_unstringify(v) as `node2;as_symbol`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;as_symbol""",
                """est2\tst2\tstring\t"KGTK"\tKGTK"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_lqstring(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:st1)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_lqstring(v) as `node2;is_lqstring`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;is_lqstring""",
                """est1\tst1\tstring\t"Franz Klammer"\t0"""]
        self.assert_literal_access_query_result(query, result)

        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:lq1)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_lqstring(v) as `node2;is_lqstring`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;is_lqstring""",
                """elq1\tlq1\tlqstring\t'hans'@de\t1"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_lqstring_text(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:lq1)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_lqstring_text(v) as `node2;text`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;text""",
                """elq1\tlq1\tlqstring\t'hans'@de\thans"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_lqstring_text_string(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:lq1)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_lqstring_text_string(v) as `node2;string`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;string""",
                """elq1\tlq1\tlqstring\t'hans'@de\t"hans\""""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_lqstring_lang(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:lq1)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_lqstring_lang(v) as `node2;lang`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;lang""",
                """elq1\tlq1\tlqstring\t'hans'@de\tde"""]
        self.assert_literal_access_query_result(query, result)

        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:lq2)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_lqstring_lang(v) as `node2;lang`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;lang""",
                """elq2\tlq2\tlqstring\t'otto'@de-bav\tde"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_lqstring_suffix(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:lq1)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_lqstring_suffix(v) as `node2;suffix`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;suffix""",
                """elq1\tlq1\tlqstring\t'hans'@de"""]
        self.assert_literal_access_query_result(query, result)

        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:lq2)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_lqstring_suffix(v) as `node2;suffix`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;suffix""",
                """elq2\tlq2\tlqstring\t'otto'@de-bav\t-bav"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_lqstring_lang_suffix(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:lq1)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_lqstring_lang_suffix(v) as `node2;lang_suffix`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;lang_suffix""",
                """elq1\tlq1\tlqstring\t'hans'@de\tde"""]
        self.assert_literal_access_query_result(query, result)

        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:lq2)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_lqstring_lang_suffix(v) as `node2;lang_suffix`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;lang_suffix""",
                """elq2\tlq2\tlqstring\t'otto'@de-bav\tde-bav"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_date(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:lq1)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_date(v) as `node2;is_date`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;is_date""",
                """elq1\tlq1\tlqstring\t'hans'@de\t0"""]
        self.assert_literal_access_query_result(query, result)

        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:d6)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_date(v) as `node2;is_date`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;is_date""",
                """ed6\td6\tdate\t^2020-10-30T02:03:57+10:30/9\t1"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_date_date(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:d6)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_date_date(v) as `node2;date`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;date""",
                """ed6\td6\tdate\t^2020-10-30T02:03:57+10:30/9\t^2020-10-30"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_date_time(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:d6)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_date_time(v) as `node2;time`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;time""",
                """ed6\td6\tdate\t^2020-10-30T02:03:57+10:30/9\t^02:03:57+10:30"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_date_and_time(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:d6)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_date_and_time(v) as `node2;date_and_time`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;date_and_time""",
                """ed6\td6\tdate\t^2020-10-30T02:03:57+10:30/9\t^2020-10-30T02:03:57+10:30"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_date_year(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:d6)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_date_year(v) as `node2;year`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;year""",
                """ed6\td6\tdate\t^2020-10-30T02:03:57+10:30/9\t2020"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_date_month(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:d6)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_date_month(v) as `node2;month`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;month""",
                """ed6\td6\tdate\t^2020-10-30T02:03:57+10:30/9\t10"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_date_day(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:d6)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_date_day(v) as `node2;day`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;day""",
                """ed6\td6\tdate\t^2020-10-30T02:03:57+10:30/9\t30"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_date_hour(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:d6)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_date_hour(v) as `node2;hour`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;hour""",
                """ed6\td6\tdate\t^2020-10-30T02:03:57+10:30/9\t2"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_date_minutes(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:d6)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_date_minutes(v) as `node2;minutes`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;minutes""",
                """ed6\td6\tdate\t^2020-10-30T02:03:57+10:30/9\t3"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_date_seconds(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:d6)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_date_seconds(v) as `node2;seconds`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;seconds""",
                """ed6\td6\tdate\t^2020-10-30T02:03:57+10:30/9\t57"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_date_zone(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:d6)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_date_zone(v) as `node2;date_zone`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;date_zone""",
                """ed6\td6\tdate\t^2020-10-30T02:03:57+10:30/9\t+10:30"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_date_zone_string(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:d6)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_date_zone_string(v) as `node2;zone_string`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;zone_string""",
                """ed6\td6\tdate\t^2020-10-30T02:03:57+10:30/9\t"+10:30\""""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_date_precision(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:d6)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_date_precision(v) as `node2;precision`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;precision""",
                """ed6\td6\tdate\t^2020-10-30T02:03:57+10:30/9\t9"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_number(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1)-[r:quantity]->(v)'
                        --return 'r, n1, r.label, v, kgtk_number(v) as `node2;is_number`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;is_number""",
                """eq1\tq1\tquantity\t0\t1""",
                """eq2\tq2\tquantity\t0.0\t1""",
                """eq3\tq3\tquantity\t+1234\t1""",
                """eq4\tq4\tquantity\t-12345.1234\t1""",
                """eq5\tq5\tquantity\t4567.12e-10\t1""",
                """eq6\tq6\tquantity\t100m\t0""",
                """eq7\tq7\tquantity\t+1.609344e03[-0.1,+0.2]m\t0""",
                """eq8\tq8\tquantity\t1.609344e03[-0.1,+0.2]Q11573\t0"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_quantity(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1)-[r:quantity]->(v)'
                        --return 'r, n1, r.label, v, kgtk_quantity(v) as `node2;is_quantity`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;is_quantity""",
                """eq1\tq1\tquantity\t0\t0""",
                """eq2\tq2\tquantity\t0.0\t0""",
                """eq3\tq3\tquantity\t+1234\t0""",
                """eq4\tq4\tquantity\t-12345.1234\t0""",
                """eq5\tq5\tquantity\t4567.12e-10\t0""",
                """eq6\tq6\tquantity\t100m\t1""",
                """eq7\tq7\tquantity\t+1.609344e03[-0.1,+0.2]m\t1""",
                """eq8\tq8\tquantity\t1.609344e03[-0.1,+0.2]Q11573\t1"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_quantity_numeral(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1)-[r:quantity]->(v)'
                        --return 'r, n1, r.label, v, kgtk_quantity_numeral(v) as `node2;numeral`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;numeral""",
                """eq1\tq1\tquantity\t0\t0""",
                """eq2\tq2\tquantity\t0.0\t0.0""",
                """eq3\tq3\tquantity\t+1234\t+1234""",
                """eq4\tq4\tquantity\t-12345.1234\t-12345.1234""",
                """eq5\tq5\tquantity\t4567.12e-10\t4567.12e-10""",
                """eq6\tq6\tquantity\t100m\t100""",
                """eq7\tq7\tquantity\t+1.609344e03[-0.1,+0.2]m\t+1.609344e03""",
                """eq8\tq8\tquantity\t1.609344e03[-0.1,+0.2]Q11573\t1.609344e03"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_quantity_numeral_string(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1)-[r:quantity]->(v)'
                        --return 'r, n1, r.label, v, kgtk_quantity_numeral_string(v) as `node2;numeral_string`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;numeral_string""",
                """eq1\tq1\tquantity\t0\t"0\"""",
                """eq2\tq2\tquantity\t0.0\t"0.0\"""",
                """eq3\tq3\tquantity\t+1234\t"+1234\"""",
                """eq4\tq4\tquantity\t-12345.1234\t"-12345.1234\"""",
                """eq5\tq5\tquantity\t4567.12e-10\t"4567.12e-10\"""",
                """eq6\tq6\tquantity\t100m\t"100\"""",
                """eq7\tq7\tquantity\t+1.609344e03[-0.1,+0.2]m\t"+1.609344e03\"""",
                """eq8\tq8\tquantity\t1.609344e03[-0.1,+0.2]Q11573\t"1.609344e03\""""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_quantity_number(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1)-[r:quantity]->(v)'
                        --return 'r, n1, r.label, v, kgtk_quantity_number(v) as `node2;number`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;number""",
                """eq1\tq1\tquantity\t0\t0""",
                """eq2\tq2\tquantity\t0.0\t0.0""",
                """eq3\tq3\tquantity\t+1234\t1234""",
                """eq4\tq4\tquantity\t-12345.1234\t-12345.1234""",
                """eq5\tq5\tquantity\t4567.12e-10\t4.56712e-07""",
                """eq6\tq6\tquantity\t100m\t100""",
                """eq7\tq7\tquantity\t+1.609344e03[-0.1,+0.2]m\t1609.344""",
                """eq8\tq8\tquantity\t1.609344e03[-0.1,+0.2]Q11573\t1609.344"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_quantity_number_int(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1)-[r:quantity]->(v)'
                        --return 'r, n1, r.label, v, kgtk_quantity_number_int(v) as `node2;number_int`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;number_int""",
                """eq1\tq1\tquantity\t0\t0""",
                """eq2\tq2\tquantity\t0.0\t0""",
                """eq3\tq3\tquantity\t+1234\t1234""",
                """eq4\tq4\tquantity\t-12345.1234\t-12345""",
                """eq5\tq5\tquantity\t4567.12e-10\t0""",
                """eq6\tq6\tquantity\t100m\t100""",
                """eq7\tq7\tquantity\t+1.609344e03[-0.1,+0.2]m\t1609""",
                """eq8\tq8\tquantity\t1.609344e03[-0.1,+0.2]Q11573\t1609"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_quantity_number_float(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1)-[r:quantity]->(v)'
                        --return 'r, n1, r.label, v, kgtk_quantity_number_float(v) as `node2;number_float`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;number_float""",
                """eq1\tq1\tquantity\t0\t0.0""",
                """eq2\tq2\tquantity\t0.0\t0.0""",
                """eq3\tq3\tquantity\t+1234\t1234.0""",
                """eq4\tq4\tquantity\t-12345.1234\t-12345.1234""",
                """eq5\tq5\tquantity\t4567.12e-10\t4.56712e-07""",
                """eq6\tq6\tquantity\t100m\t100.0""",
                """eq7\tq7\tquantity\t+1.609344e03[-0.1,+0.2]m\t1609.344""",
                """eq8\tq8\tquantity\t1.609344e03[-0.1,+0.2]Q11573\t1609.344"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_quantity_si_units(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1)-[r:quantity]->(v)'
                        --return 'r, n1, r.label, v, kgtk_quantity_si_units(v) as `node2;si_units`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;si_units""",
                """eq1\tq1\tquantity\t0\t""",
                """eq2\tq2\tquantity\t0.0\t""",
                """eq3\tq3\tquantity\t+1234\t""",
                """eq4\tq4\tquantity\t-12345.1234\t""",
                """eq5\tq5\tquantity\t4567.12e-10\t""",
                """eq6\tq6\tquantity\t100m\tm""",
                """eq7\tq7\tquantity\t+1.609344e03[-0.1,+0.2]m\tm""",
                """eq8\tq8\tquantity\t1.609344e03[-0.1,+0.2]Q11573\t"""]
        self.assert_literal_access_query_result(query, result, rowids=['eq5', 'eq6', 'eq7', 'eq8'])

    def test_kgtk_query_literal_access_kgtk_quantity_wd_units(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1)-[r:quantity]->(v)'
                        --return 'r, n1, r.label, v, kgtk_quantity_wd_units(v) as `node2;wd_units`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;wd_units""",
                """eq1\tq1\tquantity\t0\t""",
                """eq2\tq2\tquantity\t0.0\t""",
                """eq3\tq3\tquantity\t+1234\t""",
                """eq4\tq4\tquantity\t-12345.1234\t""",
                """eq5\tq5\tquantity\t4567.12e-10\t""",
                """eq6\tq6\tquantity\t100m\t""",
                """eq7\tq7\tquantity\t+1.609344e03[-0.1,+0.2]m\t""",
                """eq8\tq8\tquantity\t1.609344e03[-0.1,+0.2]Q11573\tQ11573"""]
        self.assert_literal_access_query_result(query, result, rowids=['eq5', 'eq6', 'eq7', 'eq8'])

    def test_kgtk_query_literal_access_kgtk_quantity_tolerance(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1)-[r:quantity]->(v)'
                        --return 'r, n1, r.label, v, kgtk_quantity_tolerance(v) as `node2;tolerance`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;tolerance""",
                """eq1\tq1\tquantity\t0\t""",
                """eq2\tq2\tquantity\t0.0\t""",
                """eq3\tq3\tquantity\t+1234\t""",
                """eq4\tq4\tquantity\t-12345.1234\t""",
                """eq5\tq5\tquantity\t4567.12e-10\t""",
                """eq6\tq6\tquantity\t100m\t""",
                """eq7\tq7\tquantity\t+1.609344e03[-0.1,+0.2]m\t[-0.1,+0.2]""",
                """eq8\tq8\tquantity\t1.609344e03[-0.1,+0.2]Q11573\t[-0.1,+0.2]"""]
        self.assert_literal_access_query_result(query, result, rowids=['eq5', 'eq6', 'eq7', 'eq8'])

    def test_kgtk_query_literal_access_kgtk_quantity_tolerance_string(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1)-[r:quantity]->(v)'
                        --return 'r, n1, r.label, v, kgtk_quantity_tolerance_string(v) as `node2;tolerance_string`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;tolerance_string""",
                """eq1\tq1\tquantity\t0\t""",
                """eq2\tq2\tquantity\t0.0\t""",
                """eq3\tq3\tquantity\t+1234\t""",
                """eq4\tq4\tquantity\t-12345.1234\t""",
                """eq5\tq5\tquantity\t4567.12e-10\t""",
                """eq6\tq6\tquantity\t100m\t""",
                """eq7\tq7\tquantity\t+1.609344e03[-0.1,+0.2]m\t"[-0.1,+0.2]\"""",
                """eq8\tq8\tquantity\t1.609344e03[-0.1,+0.2]Q11573\t"[-0.1,+0.2]\""""]
        self.assert_literal_access_query_result(query, result, rowids=['eq5', 'eq6', 'eq7', 'eq8'])

    def test_kgtk_query_literal_access_kgtk_quantity_low_tolerance(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1)-[r:quantity]->(v)'
                        --return 'r, n1, r.label, v, kgtk_quantity_low_tolerance(v) as `node2;low_tolerance`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;low_tolerance""",
                """eq1\tq1\tquantity\t0\t""",
                """eq2\tq2\tquantity\t0.0\t""",
                """eq3\tq3\tquantity\t+1234\t""",
                """eq4\tq4\tquantity\t-12345.1234\t""",
                """eq5\tq5\tquantity\t4567.12e-10\t""",
                """eq6\tq6\tquantity\t100m\t""",
                """eq7\tq7\tquantity\t+1.609344e03[-0.1,+0.2]m\t-0.1""",
                """eq8\tq8\tquantity\t1.609344e03[-0.1,+0.2]Q11573\t-0.1"""]
        self.assert_literal_access_query_result(query, result, rowids=['eq5', 'eq6', 'eq7', 'eq8'])

    def test_kgtk_query_literal_access_kgtk_quantity_high_tolerance(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1)-[r:quantity]->(v)'
                        --return 'r, n1, r.label, v, kgtk_quantity_high_tolerance(v) as `node2;high_tolerance`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;high_tolerance""",
                """eq1\tq1\tquantity\t0\t""",
                """eq2\tq2\tquantity\t0.0\t""",
                """eq3\tq3\tquantity\t+1234\t""",
                """eq4\tq4\tquantity\t-12345.1234\t""",
                """eq5\tq5\tquantity\t4567.12e-10\t""",
                """eq6\tq6\tquantity\t100m\t""",
                """eq7\tq7\tquantity\t+1.609344e03[-0.1,+0.2]m\t0.2""",
                """eq8\tq8\tquantity\t1.609344e03[-0.1,+0.2]Q11573\t0.2"""]
        self.assert_literal_access_query_result(query, result, rowids=['eq5', 'eq6', 'eq7', 'eq8'])

    def test_kgtk_query_literal_access_kgtk_geo_coords(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1:st1)-[r]->(v)'
                        --return 'r, n1, r.label, v, kgtk_geo_coords(v) as `node2;is_geo_coords`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;is_geo_coords""",
                """est1\tst1\tstring\t"Franz Klammer"\t0"""]
        self.assert_literal_access_query_result(query, result)

        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1)-[r:geoloc]->(v)'
                        --return 'r, n1, r.label, v, kgtk_geo_coords(v) as `node2;is_geo_coords`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;is_geo_coords""",
                """egl1\tgl1\tgeoloc\t@-42.42/69.123\t1""",
                """egl2\tgl2\tgeoloc\t@19.42/-69.123e-1\t1"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_geo_coords_lat(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1)-[r:geoloc]->(v)'
                        --return 'r, n1, r.label, v, kgtk_geo_coords_lat(v) as `node2;latitude`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;latitude""",
                """egl1\tgl1\tgeoloc\t@-42.42/69.123\t-42.42""",
                """egl2\tgl2\tgeoloc\t@19.42/-69.123e-1\t19.42"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_kgtk_geo_coords_long(self):
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1)-[r:geoloc]->(v)'
                        --return 'r, n1, r.label, v, kgtk_geo_coords_long(v) as `node2;longitude`'
                """
        result=["""id\tnode1\tlabel\tnode2\tnode2;longitude""",
                """egl1\tgl1\tgeoloc\t@-42.42/69.123\t69.123""",
                """egl2\tgl2\tgeoloc\t@19.42/-69.123e-1\t-6.9123"""]
        self.assert_literal_access_query_result(query, result)

    def test_kgtk_query_literal_access_big_numbers(self):
        """Test overflow from very long integer values and parsing of floating point constants.
        """
        query = """kgtk query -i {LITERALS} -o {OUTPUT} --graph-cache {DB}
                        --match  '(n1)-[r]->(v)' --limit 1
                        --return 'kgtk_quantity_number("+162000000000000000000000Q122922") as n1, \
                                  kgtk_quantity_number("-162000000000000000000000Q122922") as n2, \
                                  kgtk_quantity_number_int("+162000000000000000000000Q122922") as n3, \
                                  kgtk_quantity_number_int("+162000000000000000000000Q122922") as n4, \
                                  kgtk_quantity_number_float("+162000000000000000000000Q122922") as n5, \
                                  kgtk_quantity_number_float("-162000000000000000000000Q122922") as n6, \
                                  1.7976931348623157e+308 as f1, +1.7976931348623157e+308 as f2, -1.7976931348623157e+308 as f3, \
                                  2.2250738585072014e-308 as f4, +2.2250738585072014e-308 as f5, -2.2250738585072014e-308 as f6 \
                                 '
                """
        # TO DO: we might need to make this test more tolerant to work around any floating point precision issues:
        result=["""n1\tn2\tn3\tn4\tn5\tn6\tf1\tf2\tf3\tf4\tf5\tf6""",
                """1.62e+23\t-1.62e+23\t9223372036854775807\t9223372036854775807\t1.62e+23\t-1.62e+23"""
                + """\t1.7976931348623157e+308\t1.7976931348623157e+308\t-1.7976931348623157e+308\t"""
                + """2.2250738585072014e-308\t2.2250738585072014e-308\t-2.2250738585072014e-308"""]
        self.assert_literal_access_query_result(query, result)



"""
Cypher grammar (derived from ruruki.parsers.cypher_parser.py)
https://s3.amazonaws.com/artifacts.opencypher.org/cypher.ebnf
https://s3.amazonaws.com/artifacts.opencypher.org/openCypher9.pdf
"""

# NOTES:
# - this is patterned after the cypher.ebnf grammar which seems to be a Neo product
#   and which is somewhat different from what is in openCypher9.pdf, e.g., the
#   Open Cypher examples use multiple MATCH clauses instead of comma chains
# - the latest cypher.ebnf is copyrighted 2020 while the grammar below was derived
#   from some 2016 version which might have differed slightly
# - the cypher.ebnf in this directory has a August 11, 2020, 7:06:53 AM PDT page mod date
# - cypher.ebnf and the version below only seems to support only a single triple in a
#   where clause, e.g., "where not (a)-[:P23]->(b)" as opposed to arbitrary chains;
#   is this a grammar bug or intentional restriction?
# - we should eliminate all parts of the grammar we do not support such as the various
#   update operations; better is to wrap relevant parts with ["Unsupported" ....]
#   directly in the grammar, thus allowing us later to gradually support them if we want
# - Cypher is a registered trademark, and we are not allowed to call what we build
#   "Cypher" - see the copyright on the grammar

# - EBNF:
#   - cypher.ebnf uses Extended BNF for which ISO/IEC 14977:1996 is a (proposed) standard
#   - we downloaded the standard from here to review some quirks:
#     - https://standards.iso.org/ittf/PubliclyAvailableStandards/s026153_ISO_IEC_14977_1996(E).zip
#   - meta characters: [] means 0 or 1 times, {} means 0 or more time, - is the exception symbol
#     that means everything in the first clause not matched by the second exception clause,
#     finally {...}-, means one time or more, since it has an empty exception - that one was tricky
#   - we might be able to automatically convert the EBNF grammar into a Parsley grammar that returns
#     a structured parse tree as a result, which we'd have to then further normalize; the problem
#     with the current scheme is that it uses a custom rewrite of EBNF into Parsley which is a bit
#     "wooly" and is harder to adapt when the underlying Cypher grammar changes
#   - we basically want each keyword to return itself, and each toplevel alternative of each rule
#     return the rule's name with the arguments of that disjunction.  For example (or something like that):
#     - EBNF:     OrExpression = XorExpression, { SP, (O,R), SP, XorExpression } ;
#     - Parsley:  OrExpression = XorExpression:x1 (WS (O R) -> "OR" WS XorExpression)*:x2 -> ["OrExpression", x1, x2]


KYPHER_GRAMMAR = r"""
    Kypher = WS Statement:s (WS ';')? WS -> s

    Statement = Query

    Query = RegularQuery

    RegularQuery = SingleQuery:sq SP U N I O N SP A L L SP RegularQuery:rq -> ["UnionAll", sq, rq]
                 | SingleQuery:sq SP U N I O N SP RegularQuery:rq -> ["Union", sq, rq]
                 | SingleQuery:sq

    # SingleQuery = Clause:head (WS Clause)*:tail -> ["SingleQuery", [head] + tail]

    # Clause = Match
    #        | Unwind
    #        | Merge
    #        | Create
    #        | Set
    #        | Delete
    #        | Remove
    #        | With
    #        | Return

    SingleQuery = Match?:m WS With?:w WS Return:r -> ["SingleQuery", m, w, r]

    # TODO: Not sure if I need to handle optional !!
    Match = (O P T I O N A L SP)? M A T C H WS Pattern:p (WS Where)?:w -> ["Match", p, w]

    Unwind = U N W I N D WS Expression:ex SP A S SP Variable:v -> ["Unwind", ex, v]

    Merge = M E R G E WS PatternPart:head (SP MergeAction)*:tail -> ["Merge", [head] + tail]

    MergeAction = O N SP M A T C H SP Set:s -> ["MergeActionMatch", s]
                | O N SP C R E A T E SP Set:s -> ["MergeActionCreate", s]

    Create = C R E A T E WS Pattern:p -> ["Create", p]

    Set = S E T SP SetItem:head (WS ',' WS SetItem)*:tail -> ["Set", [head] + tail]

    SetItem = PropertyExpression:pex '=' Expression:ex -> ["SetItemPropertyExpression", pex, ex]
            | Variable:v '=' Expression:ex -> ["SetItem", v, ex]
            | Variable:v '+=' Expression:ex -> ["SetItem", v, ex]
            | Variable:v NodeLabels:ex -> ["SetItem", v, ex]

    Delete = (D E T A C H SP)? D E L E T E SP Expression:head (',' WS Expression )*:tail -> ["Delete", [head] + tail]

    Remove = R E M O V E SP RemoveItem:head (WS ',' WS RemoveItem)*:tail -> ["Remove", [head] + tail]

    RemoveItem = Variable:v NodeLabels:nl -> ["RemoveItemVar", v, nl]
                | PropertyExpression:p -> ["RemoveItemPe", p]

    With = W I T H (SP D I S T I N C T)?:d SP ReturnBody:rb (Where)?:w -> ["With", d, rb, w]

    Return = R E T U R N (SP D I S T I N C T)?:d SP ReturnBody:rb -> ["Return", d, rb]

    ReturnBody = ReturnItems:ri (SP Order)?:o (SP Skip)?:s (SP Limit)?:l -> ["ReturnBody", ri, o, s, l]

    ReturnItems = ('*' | ReturnItem):head
                (WS ',' WS ReturnItem )*:tail -> ["ReturnItems", [head] + tail]

    ReturnItem = Expression:ex SP A S SP SymbolicName:s -> ["ReturnItem", ex, s]
               | Expression:ex -> ["ReturnItem", ex, None]

    Order =  O R D E R SP B Y SP SortItem:head (WS ',' WS SortItem)*:tail -> ["Order", [head] + tail]

    Skip =  S K I P SP Expression:ex -> ["Skip", ex]

    Limit =  L I M I T SP Expression:ex -> ["Limit", ex]

    SortItem = Expression:ex (SP D E S C E N D I N G | SP D E S C) -> ["sort", ex, "desc"]
             | Expression:ex (SP A S C E N D I N G | SP A S C)? -> ["sort", ex, "asc"]

    Where = W H E R E SP Expression:ex -> ["Where", ex]

    Pattern = PatternPart:head (',' WS PatternPart)*:tail -> [head] + tail

    PatternPart = (Variable:v WS '=' WS AnonymousPatternPart:ap) -> ["PatternPart", v, ap]
                | (Variable:v ':' WS AnonymousPatternPart:ap) -> ["GraphPatternPart", v, ap]
                | AnonymousPatternPart:ap -> ["PatternPart", None, ap]

    AnonymousPatternPart = PatternElement

    PatternElement = (
                        NodePattern:np
                        (WS PatternElementChain)*:pec
                    ) -> ["PatternElement", np, pec]
                    | '(' PatternElement:pe ')' -> pe

    NodePattern = '(' WS
                 (
                    SymbolicName:s WS -> s
                 )?:s
                 (
                     NodeLabels:nl WS -> nl
                 )?:nl
                 (
                     Properties:p WS -> p
                 )?:p
                ')' -> ["NodePattern", s, nl, p]

    PatternElementChain = RelationshipPattern:rp WS NodePattern:np -> ["PatternElementChain", rp, np]

    RelationshipPattern = LeftArrowHead?:la WS Dash WS RelationshipDetail?:rd WS Dash WS RightArrowHead?:ra -> ["RelationshipsPattern", la, rd, ra]

    # TO DO: fix WS handling here to be more liberal and match what we do for NodePattern:
    RelationshipDetail = '['
                      Variable?:v
                      '?'?:q
                      RelationshipTypes?:rt
                      ('*' RangeLiteral)?:rl WS
                      Properties?:p
                      ']' -> ["RelationshipDetail", v, q, rt, rl, p]

    Properties = MapLiteral
               | Parameter

    RelationshipTypes = ':' RelTypeName:head (WS '|' ':'? WS RelTypeName)*:tail -> ["RelationshipTypes", head] + tail

    NodeLabels = NodeLabel:head (WS NodeLabel)*:tail -> [head] + tail

    NodeLabel = ':' LabelName:n -> ["NodeLabel", n]

    RangeLiteral = (WS IntegerLiteral)?:start WS ('..' WS IntegerLiteral)?:stop WS -> slice(start, stop)

    LabelName = SymbolicName

    RelTypeName = SymbolicName

    Expression = Expression12

    Expression12 = Expression11:ex1 SP O R SP Expression12:ex2 -> ["or", ex1, ex2]
                 | Expression11

    Expression11 = Expression10:ex1 SP X O R SP Expression11:ex2 -> ["xor", ex1, ex2]
                 | Expression10

    Expression10 = Expression9:ex1 SP A N D SP Expression10:ex2 -> ["and", ex1, ex2]
                 | Expression9

    Expression9 = N O T SP Expression9:ex -> ["not", ex]
                | Expression8

    Expression8 = Expression7:ex1 WS '='  WS Expression8:ex2 -> ["eq",  ex1, ex2]
                | Expression7:ex1 WS '<>' WS Expression8:ex2 -> ["neq", ex1, ex2]
                | Expression7:ex1 WS '!=' WS Expression8:ex2 -> ["neq", ex1, ex2]
                | Expression7:ex1 WS '<'  WS Expression8:ex2 -> ["lt",  ex1, ex2]
                | Expression7:ex1 WS '>'  WS Expression8:ex2 -> ["gt",  ex1, ex2]
                | Expression7:ex1 WS '<=' WS Expression8:ex2 -> ["lte", ex1, ex2]
                | Expression7:ex1 WS '>=' WS Expression8:ex2 -> ["gte", ex1, ex2]
                | Expression7

    Expression7 = Expression6:ex1 WS '+' WS Expression7:ex2 -> ["add", ex1, ex2]
                | Expression6:ex1 WS '-' WS Expression7:ex2 -> ["sub", ex1, ex2]
                | Expression6

    Expression6 = Expression5:ex1 WS '*' WS Expression6:ex2 -> ["multi", ex1, ex2]
                | Expression5:ex1 WS '/' WS Expression6:ex2 -> ["div",   ex1, ex2]
                | Expression5:ex1 WS '%' WS Expression6:ex2 -> ["mod",   ex1, ex2]
                | Expression5

    Expression5 = Expression4:ex1 WS '^' WS Expression5:ex2 -> ["hat", ex1, ex2]
                | Expression4

    Expression4 = '+' WS Expression4
                | '-' WS Expression4:ex -> ["minus", ex]
                | Expression3

    Expression3 = Expression2:ex1
                  (
                    WS '[' Expression:prop_name ']' -> ["PropertyLookup", prop_name]
                    | WS '[' Expression?:start '..' Expression?:end ']' -> ["slice", start, end]
                    | (
                        WS '=~' -> "regex"
                        | SP I N -> "in"
                        | SP S T A R T S SP W I T H -> "starts_with"
                        | SP E N D S SP W I T H  -> "ends_with"
                        | SP C O N T A I N S  -> "contains"
                    ):operator WS Expression2:ex2 -> [operator, ex2]
                    | SP I S SP N U L L  -> ["is_null"]
                    | SP I S SP N O T SP N U L L -> ["is_not_null"]
                  )+:c -> ["Expression3", ex1, c]
                  | Expression2

    Expression2 = Atom:a (PropertyLookup | NodeLabels)+:c -> ["Expression2", a, c]
                  | Atom

    Atom = NumberLiteral
         | StringLiteral
         | Parameter
         | T R U E -> ["Literal", True]
         | F A L S E -> ["Literal", False]
         | N U L L -> ["Literal", None]
         | CaseExpression
         | C O U N T '(' '*' ')' -> ["count *"]
         | MapLiteral
         | ListComprehension
         | '['
                (
                    WS Expression:head WS
                    (',' WS Expression:item WS -> item
                    )*:tail -> [head] + tail
                    |
                    -> []
                ):ex
            ']' -> ["List", ex]
         | F I L T E R WS '(' WS FilterExpression:fex WS ')' -> ["Filter", fex]
         | E X T R A C T WS '(' WS FilterExpression:fex WS (WS '|' Expression)?:ex ')' -> ["Extract", fex, ex]
         | A L L WS '(' WS FilterExpression:fex WS ')' -> ["All", fex]
         | A N Y WS '(' WS FilterExpression:fex WS ')' -> ["Any", fex]
         | N O N E WS '(' WS FilterExpression:fex WS ')' -> ["None", fex]
         | S I N G L E WS '(' WS FilterExpression:fex WS ')' -> ["Single", fex]
         | RelationshipsPattern
         | GraphRelationshipsPattern
         | parenthesizedExpression
         | FunctionInvocation
         | Variable

    parenthesizedExpression = '(' WS Expression:ex WS ')' -> ex

    RelationshipsPattern = NodePattern:np (WS PatternElementChain)?:pec -> ["RelationshipsPattern", np, pec]
    
    GraphRelationshipsPattern = Variable:v ':' WS NodePattern:np (WS PatternElementChain)?:pec -> ["GraphRelationshipsPattern", v, np, pec]

    FilterExpression = IdInColl:i (WS Where)?:w -> ["FilterExpression", i, w]

    IdInColl = Variable:v SP I N SP Expression:ex -> ["IdInColl", v, ex]

    FunctionInvocation = FunctionName:func
                        WS '(' WS
                        (D I S T I N C T WS -> "distinct")?:distinct
                        (
                            Expression:head
                            (
                                ',' WS Expression
                            )*:tail -> [head] + tail
                        | -> []
                        ):args
                        WS ')' -> ["call", func, distinct, args]

    FunctionName = SymbolicName

    ListComprehension = '[' FilterExpression:fex (WS '|' Expression)?:ex ']' -> ["ListComprehension", fex, ex]

    # PropertyLookup = WS '.' WS ((PropertyKeyName ('?' | '!')) | PropertyKeyName)
    PropertyLookup = WS '.' WS PropertyKeyName:n -> ["PropertyLookup", n]

    CaseExpression =
                     C A S E WS
                     (Expression)?:ex
                     (WS CaseAlternatives)+:cas 
                     (WS E L S E WS Expression)?:el
                     WS E N D
                     -> ["Case", ex, cas, el]

    CaseAlternatives = W H E N WS Expression:ex1 WS T H E N WS Expression:ex2 -> [ex1, ex2]

    Variable = SymbolicName:s -> ["Variable", s]

    StringLiteral = (
                  '"' (~('"'|'\\') anything | EscapedChar)*:cs '"' -> "".join(cs)
                  | "'" (~("'"|'\\') anything | EscapedChar)*:cs "'" -> "".join(cs)
                  ):l -> ["Literal", l]

    EscapedChar = '\\'
                ('\\' -> '\\'
                | "'" -> "'"
                | '"' -> '"'
                | N -> '\n'
                | R -> '\r'
                | T -> '\t'
                | '_' -> '_'
                | '%' -> '%'
                )

    NumberLiteral = (
                  DoubleLiteral
                  | IntegerLiteral
                  ):l -> ["Literal", l]

    MapLiteral = '{' WS
                 (
                    (
                        PropertyKeyName:k WS ':' WS Expression:v -> (k, v)
                    ):head WS
                    (
                        ',' WS PropertyKeyName:k WS ':' WS Expression:v WS -> (k, v)
                    )*:tail -> [head] + tail
                 | -> []):pairs
                '}' -> ["Literal", dict(pairs)]

    Parameter = '$' (SymbolicName | DecimalInteger):p -> ["Parameter", p]

    PropertyExpression = Atom:a (WS PropertyLookup)*:opts -> ["Expression", a, opts]

    PropertyKeyName = SymbolicName

    IntegerLiteral = HexInteger
                   | OctalInteger
                   | DecimalInteger

    OctalDigit = ~('8'|'9') digit

    OctalInteger = '0' <OctalDigit+>:ds -> int(ds, 8)

    HexDigit = digit | A | B | C | D | E | F

    HexInteger = '0' X <HexDigit+>:ds -> int(ds, 16)

    DecimalInteger = <digit+>:ds -> int(ds)

    DoubleLiteral = ExponentDecimalReal
                  | RegularDecimalReal

    ExponentDecimalReal = <(RegularDecimalReal | DecimalInteger) E ('+' | '-')? DecimalInteger>:ds -> float(ds)

    RegularDecimalReal = (<digit+ '.' digit+> | <digit+ '.'> | <'.' digit+>):ds -> float(ds)

    SymbolicName = UnescapedSymbolicName
                 | EscapedSymbolicName

    UnescapedSymbolicName = <letter ('_' | letterOrDigit)*>

    EscapedSymbolicName = '`' (~'`' anything | "``" -> '`')*:cs '`' -> "".join(cs)

    WS = whitespace*

    SP = whitespace+

    whitespace = ' '
               | '\t'
               | '\n'
               | Comment

    Comment = "/*" (~"*/" anything)* "*/"
            | "//" (~('\r'|'\n') anything)* '\r'? ('\n'|end)

    LeftArrowHead = '<'

    RightArrowHead = '>'

    Dash = '-'

    A = 'A' | 'a'

    B = 'B' | 'b'

    C = 'C' | 'c'

    D = 'D' | 'd'

    E = 'E' | 'e'

    F = 'F' | 'f'

    G = 'G' | 'g'

    H = 'H' | 'h'

    I = 'I' | 'i'

    K = 'K' | 'k'

    L = 'L' | 'l'

    M = 'M' | 'm'

    N = 'N' | 'n'

    O = 'O' | 'o'

    P = 'P' | 'p'

    R = 'R' | 'r'

    S = 'S' | 's'

    T = 'T' | 't'

    U = 'U' | 'u'

    V = 'V' | 'v'

    W = 'W' | 'w'

    X = 'X' | 'x'

    Y = 'Y' | 'y'

    Z = 'Z' | 'z'
    """



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE.txt file in the root directory of this source tree.

import argparse
from typing import Iterable, TextIO
import logging
from torchbiggraph.checkpoint_manager import CheckpointManager
from torchbiggraph.config import ConfigFileLoader, ConfigSchema
from torchbiggraph.graph_storages import (
    ENTITY_STORAGES,
    RELATION_TYPE_STORAGES,
    AbstractEntityStorage,
    AbstractRelationTypeStorage,
)
from torchbiggraph.model import MultiRelationEmbedder, make_model


def write(outf: TextIO, key: Iterable[str], value: Iterable[float]) -> None:
    outf.write("%s\t%s\n" % ("\t".join(key), "\t".join("%.9f" % x for x in value)))


def make_tsv(
    config: ConfigSchema, entities_tf: TextIO, relation_types_tf: TextIO
) -> None:
    logging.info("Loading relation types and entities...")
    entity_storage = ENTITY_STORAGES.make_instance(config.entity_path)
    relation_type_storage = RELATION_TYPE_STORAGES.make_instance(config.entity_path)

    logging.info("Initializing model...")
    model = make_model(config)

    logging.info("Loading model check point...")
    checkpoint_manager = CheckpointManager(config.checkpoint_path)
    state_dict, _ = checkpoint_manager.read_model()
    if state_dict is not None:
        model.load_state_dict(state_dict, strict=False)

    make_tsv_for_entities(model, checkpoint_manager, entity_storage, entities_tf)
    if config.relations[0].operator != 'linear': # when the operator is linear, the relations type will raise an error
        make_tsv_for_relation_types(model, relation_type_storage, relation_types_tf)


def make_tsv_for_entities(
    model: MultiRelationEmbedder,
    checkpoint_manager: CheckpointManager,
    entity_storage: AbstractEntityStorage,
    entities_tf: TextIO,
) -> None:
    logging.info("Writing entity embeddings...")
    for ent_t_name, ent_t_config in model.entities.items():
        for partition in range(ent_t_config.num_partitions):
            logging.info(
                f"Reading embeddings for entity type {ent_t_name} partition "
                f"{partition} from checkpoint..."
            )
            entities = entity_storage.load_names(ent_t_name, partition)
            embeddings, _ = checkpoint_manager.read(ent_t_name, partition)

            if model.global_embs is not None:
                embeddings += model.global_embs[model.EMB_PREFIX + ent_t_name]

            logging.info(
                f"Writing embeddings for entity type {ent_t_name} partition "
                f"{partition} to output file..."
            )
            for ix in range(len(embeddings)):
                write(entities_tf, (entities[ix],), embeddings[ix])
                if (ix + 1) % 5000 == 0:
                    logging.info(f"- Processed {ix+1}/{len(embeddings)} entities so far...")
            logging.info(f"- Processed all {len(embeddings)} entities")

    entities_output_filename = getattr(entities_tf, "name", "the output file")
    logging.info(f"Done exporting entity data to {entities_output_filename}")


def make_tsv_for_relation_types(
    model: MultiRelationEmbedder,
    relation_type_storage: AbstractRelationTypeStorage,
    relation_types_tf: TextIO,
) -> None:
    logging.info("Writing relation type parameters...")
    relation_types = relation_type_storage.load_names()
    if model.num_dynamic_rels > 0:
        (rel_t_config,) = model.relations
        op_name = rel_t_config.operator
        (lhs_operator,) = model.lhs_operators
        (rhs_operator,) = model.rhs_operators
        for side, operator in [("lhs", lhs_operator), ("rhs", rhs_operator)]:
            for param_name, all_params in operator.named_parameters():
                for rel_t_name, param in zip(relation_types, all_params):
                    shape = "x".join(f"{d}" for d in param.shape)
                    write(
                        relation_types_tf,
                        (rel_t_name, side, op_name, param_name, shape),
                        param,
                    )
    else:
        for rel_t_name, rel_t_config, operator in zip(
            relation_types, model.relations, model.rhs_operators
        ):
            if rel_t_name != rel_t_config.name:
                raise ValueError(
                    f"Mismatch in relations names: got {rel_t_name} in the "
                    f"dictionary and {rel_t_config.name} in the config."
                )
            op_name = rel_t_config.operator
            for param_name, param in operator.named_parameters():
                shape = "x".join(f"{d}" for d in param.shape)
                write(
                    relation_types_tf,
                    (rel_t_name, "rhs", op_name, param_name, shape),
                    param,
                )

    relation_types_output_filename = getattr(
        relation_types_tf, "name", "the output file"
    )
    logging.info(f"Done exporting relation type data to {relation_types_output_filename}")


def main():
    config_help = "\n\nConfig parameters:\n\n" + "\n".join(ConfigSchema.help())
    parser = argparse.ArgumentParser(
        epilog=config_help,
        # Needed to preserve line wraps in epilog.
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument("config", help="Path to config file")
    parser.add_argument("-p", "--param", action="append", nargs="*")
    parser.add_argument("--entities-output", required=True)
    parser.add_argument("--relation-types-output", required=True)
    opt = parser.parse_args()

    loader = ConfigFileLoader()
    config = loader.load_config(opt.config, opt.param)

    with open(opt.entities_output, "xt") as entities_tf, open(
        opt.relation_types_output, "xt"
    ) as relation_types_tf:
        make_tsv(config, entities_tf, relation_types_tf)


if __name__ == "__main__":
    main()


import os
import sys
import io
import sh # type: ignore
import tempfile
import typing

from kgtk.exceptions import KGTKException
from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Concatenate any mixture of plain or gzip/bzip2/xz-compressed files'
    }

def add_arguments(parser: KGTKArgumentParser):
    parser.add_argument('-o', '--out', default=None, dest='output',
                        help='output file to write to, otherwise output goes to stdout')
    parser.add_argument('--gz', '--gzip', action='store_true', dest='gz',
                        help='compress result with gzip')
    parser.add_argument('--bz2', '--bzip2', action='store_true', dest='bz2',
                        help='compress result with bzip2')
    parser.add_argument('--xz', action='store_true', dest='xz',
                        help='compress result with xz')

    # parser.add_argument("inputs", metavar="INPUT", nargs="*", action="store",
    #                    help="input files to process, if empty or `-' read from stdin")
    parser.add_input_file(positional=True, allow_list=True, dest="input_files")


### general command utilities (some of these should make it into a more central location):

tmp_dir = '/tmp'       # this should be configurable
kgtk_encoding = 'utf8' # this should be configurable

def make_temp_file(prefix='kgtk.'):
    return tempfile.mkstemp(dir=tmp_dir, prefix=prefix)[1]

def get_buf_sizes(input=None, output=None, _tty_out=True, _piped=False):
    """Determine stream buffer sizes to use.  Since sh has complex rules for this depending on
    what streams are used and flags are set, we simply try this here and see if it lets us do it.
    We want to make sure to use large output buffers whenever possible for speed.
    This should probably go into cli_entry.py.
    """
    in_bufsize = 2**16
    out_bufsize = in_bufsize
    try:
        sh.ls.bake(_in=input, _in_bufsize=in_bufsize)
    except:
        in_bufsize = None
    try:
        sh.ls.bake(_out=output, _out_bufsize=out_bufsize, _tty_out=_tty_out, _piped=_piped)
    except:
        out_bufsize = None
    return in_bufsize, out_bufsize

def get_stream_buffer_size(stream):
    """Determine the buffer size used for `stream' which will either be the underlying file
    descriptor's block size or a default buffer size.  If we read ahead, we need to read in
    multiples of this amount, otherwise a later read on an os.dup-ed file descriptor (as
    generated by sh) will start at the next block boundary and lose data.
    """
    # TO DO: simplify the whole scheme, this is getting way too hairy
    try:
        block_size = os.stat(stream.fileno()).st_blksize
    except:
        block_size = io.DEFAULT_BUFFER_SIZE
    #sys.stderr.write('get_stream_buffer_size: %d\n' % block_size)
    return block_size

def get_stream_header(stream, preserve=False):
    """Utility to access header information from a stream (usually stdin) without losing
    the ability to pass the rest or (if `preserve' is True) all of `stream' to other consumers.
    Reads `n' `unit's from stream (or fewer if `stream' doesn't have that much content) and
    returns the result as a byte sequence.  If `preserve' the second return value is a `sh'
    command sequence that can be piped into another sh-command with the full stream content.
    """
    header = io.BytesIO()
    header.write(stream.read(get_stream_buffer_size(stream)))
    if preserve:
        # we need to pass in the header with a temporary file which will be deleted when `cat' terminates.
        # alternatively to this scheme, we could create some kind of concatenated stream class:
        temp = make_temp_file('kgtk-header.')
        with open(temp, 'wb') as out:
            out.write(header.getvalue())
        cleanup = lambda cmd, status, exit_code: sh.rm('-f', temp)
        in_bufsize, out_bufsize = get_buf_sizes(input=stream, _tty_out=False, _piped=True)
        return header.getvalue(), [sh.cat.bake(temp, '-', _in=stream, _in_bufsize=in_bufsize, _piped=True, _done=cleanup)]
    else:
        return header.getvalue()

def run_sh_commands(commands, debug=False):
    """Run a single or list of prebaked sh `commands', compose them with pipes when they
    are marked with piped=True or bg=True.  Return the last run command which can be used
    to access the final exit_code and other state.
    """
    if not hasattr(commands, "__iter__"):
        commands = [commands]
    if debug:
        sys.stderr.write('zconcat.run_sh_commands: %s\n' % commands)
    piped_output = None
    last_cmd = None
    for cmd in commands:
        if piped_output is not None:
            piped_output = cmd(piped_output)
        else:
            piped_output = cmd()
        last_cmd = piped_output
        # TO DO: improve this with a more explicit directive in case we don't use piped or bg:
        if not (cmd._partial_call_args.get('piped', False) or cmd._partial_call_args.get('bg', False)):
            piped_output = None
    return last_cmd

def determine_file_type(file):
    """Determine if `file' is compressed and if so how, and return file and its associated type.
    `file' needs to be a file name or a stream containing enough information to determine its type.
    """
    if isinstance(file, str):
        file_type = sh.file(file, '--brief').stdout
    else:
        file_type = sh.file('-', '--brief', _in=file).stdout
    # tricky: we get a byte sequence here which we have to decode into a string:
    return file_type.split()[0].lower().decode()

compression_type_table = {
    # we keep these strings, so we don't require the commands to be available during loading:
    'gzip':  {'cat': 'zcat',  'compress': 'gzip'},
    'bzip2': {'cat': 'bzcat', 'compress': 'bzip2'},
    'xz':    {'cat': 'xzcat', 'compress': 'xz'},
    'text':  {'cat': 'cat'},
}

def get_cat_command(file_type):
    """Determine a `cat' command based on a `file_type' determined by `determine_file_type'.
    """
    catcmd = compression_type_table.get(file_type, {}).get('cat', 'cat')
    # now return the equivalent of sh.cat, etc:
    return getattr(sh, catcmd)

def get_compress_command(file_type):
    """Return compress command to run based on target `file_type'.
    """
    compress = compression_type_table.get(file_type, {}).get('compress')
    return compress and getattr(sh, compress) or None


### zconcat implementation:

def compress_switch_to_file_type(gz=False, bz2=False, xz=False):
    """Return compressed target file_type based on supplied switches.
    """
    return (gz and 'gzip') or (bz2 and 'bzip2') or (xz and 'xz') or 'text'

def build_command_1(input=None, output=None, gz=False, bz2=False, xz=False, _piped=False, _out_mode='wb'):
    """Build a zconcat sh command pipe for a single `input'.
    If `_piped' is True, configure the last command to ignore `output' and write to a pipe
    in which case this can be used to feed into the input of another command (e.g., sort).
    `_out_mode' controls whether an `output' file will be truncated or appened to.
    """
    input = input or '-'
    output = (not _piped and (output or sys.stdout.buffer)) or None
    outfile = None
    if isinstance(output, str):
        outfile = open(output, _out_mode)
        output = outfile
    compress = get_compress_command(compress_switch_to_file_type(gz, bz2, xz))
    in_bufsize, out_bufsize = get_buf_sizes(output=output, _tty_out=not compress, _piped=_piped)
    
    if input == '-':
        # process input piped in from stdin, possibly compressed in different ways:
        input = sys.stdin.buffer
        header, input = get_stream_header(input, preserve=True)
        file_type = determine_file_type(io.BytesIO(header))
        catcmd = get_cat_command(file_type)
        cleanup = lambda cmd, status, exit_code: outfile and outfile.close()
        if compress is not None:
            return input + [
                catcmd.bake(_piped=True),
                compress.bake('-c', _out=output, _out_bufsize=out_bufsize, _tty_out=False, _done=cleanup, _piped=_piped)
            ]
        else:
            return input + [
                catcmd.bake(_out=output, _out_bufsize=out_bufsize, _done=cleanup, _piped=_piped)
            ]
    else:
        # process a regular named file, possibly compressed in different ways:
        file_type = determine_file_type(input)
        catcmd = get_cat_command(file_type)
        cleanup = lambda cmd, status, exit_code: outfile and outfile.close()
        if compress is not None:
            return [
                catcmd.bake(input, _piped=True),
                compress.bake('-c', _out=output, _out_bufsize=out_bufsize, _tty_out=False, _done=cleanup, _piped=_piped)
            ]
        else:
            return [
                catcmd.bake(input, _out=output, _out_bufsize=out_bufsize, _done=cleanup, _piped=_piped)
            ]

def build_command(inputs=[], output=None, gz=False, bz2=False, xz=False):
    """Build a zconcat sh command pipe for the provided `inputs' and switches.
    """
    if len(inputs) == 0:
        inputs.append('-')
    command = []
    out_mode='wb'
    for inp in inputs:
        command.extend(build_command_1(input=inp, output=output, gz=gz, bz2=bz2, xz=xz, _out_mode=out_mode))
        out_mode='ab'
    return command

# def run(inputs=[], output=None, gz=False, bz2=False, xz=False, _debug=False):
def run(input_files: KGTKFiles, output=None, gz=False, bz2=False, xz=False, _debug=False):
    """Run zconcat according to the provided command-line arguments.
    """
    # TO DO: figure out how to properly access shared --debug option
    try:
        inputs: typing.List[str] = [str(input_file) for input_file in KGTKArgumentParser.get_input_file_list(input_files)]
        commands = build_command(inputs=inputs, output=output, gz=gz, bz2=bz2, xz=xz)
        return run_sh_commands(commands, debug=_debug).exit_code
    except sh.SignalException_SIGPIPE:
        # cleanup in case we piped and terminated prematurely:
        sys.stdout.flush()
    except Exception as e:
        #import traceback
        #traceback.print_tb(sys.exc_info()[2], 10)
        raise KGTKException('INTERNAL ERROR: ' + str(e) + '\n')

"""
# Examples:

> echo hello | kgtk zconcat
hello

> cat <<EOF > /tmp/file1
line1
line2
EOF
> cat <<EOF > /tmp/file2
line3
line4
EOF

> echo hello | kgtk ticker -i / zconcat --gz -o /tmp/out.gz /tmp/file1 - /tmp/file2
> 
> bzip2 /tmp/file1
> echo hello-again | kgtk zconcat /tmp/out.gz - /tmp/file1.bz2 
line1
line2
hello
>2020-04-01 15:57:48.750904
line3
line4
hello-again
line1
line2

> cat /tmp/file1.bz2 | kgtk zconcat
line1
line2

> cat /tmp/out.gz | kgtk zconcat
line1
line2
hello
>2020-04-02 18:36:40.000507
line3
line4

> cat /tmp/out.gz | kgtk zconcat | head -4
line1
line2
hello
>2020-04-07 14:04:15.962612

# speed test on a large 2GB compressed file:
> time kgtk zconcat -o /tmp/nodes-v2.csv.gz --gz /data/kgtk/wikidata/run1/nodes-v2.csv.gz
284.356u 6.955s 4:00.00 121.3%	0+0k 0+4063472io 0pf+0w
# elapsed time is the same as doing it directly in the shell:
> date; zcat /data/kgtk/wikidata/run1/nodes-v2.csv.gz | gzip -c > /tmp/nodes-v2.csv.gz; date
Tue 07 Apr 2020 10:39:50 AM PDT
Tue 07 Apr 2020 10:43:49 AM PDT
"""



import shutil
import unittest
import tempfile
import pandas as pd
from kgtk.cli_entry import cli_entry


class TestKGTKFilter(unittest.TestCase):
    def setUp(self) -> None:
        self.file_path = 'data/sample_kgtk_edge_file.tsv'
        self.temp_dir = tempfile.mkdtemp()
        self.df = pd.read_csv(self.file_path, sep='\t')

    def tearDown(self) -> None:
        shutil.rmtree(self.temp_dir)

    def test_kgtk_ifexists(self):
        Q47158_path = 'data/sample_kgtk_edge_Q47158.tsv'
        cli_entry("kgtk", "ifexists", "-i", self.file_path, "--filter-on", Q47158_path, "-o",
                  f'{self.temp_dir}/Q47158.tsv', "--input-keys", "node1", "--filter-keys", "node1", "--show-option",
                  "--verbose")

        df = pd.read_csv(f'{self.temp_dir}/Q47158.tsv', sep='\t')

        self.assertEqual(len(df), 118)

    def test_kgtk_ifexists_mode_none(self):
        Q47158_path = 'data/Q47158_non_edge.tsv'
        cli_entry("kgtk", "ifexists", "-i", self.file_path, "--filter-on", Q47158_path, "-o",
                  f'{self.temp_dir}/Q47158.tsv', "--input-keys", "node1", "--filter-keys", "heading", "--mode", "NONE",
                  "--verbose")

        df = pd.read_csv(f'{self.temp_dir}/Q47158.tsv', sep='\t')

        self.assertEqual(len(df), 118)

    def test_kgtk_ifexists_preserve_order(self):
        Q47158_path = 'data/Q47158_non_edge.tsv'
        cli_entry("kgtk", "ifexists", "-i", self.file_path, "--filter-on", Q47158_path, "-o",
                  f'{self.temp_dir}/Q47158.tsv', "--input-keys", "node1", "--filter-keys", "heading", "--mode", "NONE",
                  "--preserve-order", "--cache-input", "--verbose")

        f = open(f'{self.temp_dir}/Q47158.tsv')
        lines = f.readlines()
        # strip off the \n
        self.assertEqual(lines[1].strip(), "Q47158-P1014-1	Q47158	P1014	300391257	normal")
        self.assertEqual(lines[len(lines) - 1].replace('\n', ''),
                         "Q47158-wikipedia_sitelink-9	Q47158	wikipedia_sitelink	http://et.wikipedia.org/wiki/Triias	")
        f.close()

    def test_kgtk_ifexists_not_preserve_order(self):
        Q47158_path = 'data/Q47158_non_edge.tsv'
        cli_entry("kgtk", "ifexists", "-i", self.file_path, "--filter-on", Q47158_path, "-o",
                  f'{self.temp_dir}/Q47158.tsv', "--input-keys", "node1", "--filter-keys", "heading", "--mode", "NONE",
                  "--cache-input", "--verbose")

        f = open(f'{self.temp_dir}/Q47158.tsv')
        lines = f.readlines()
        # strip off the \n
        self.assertEqual(lines[1].strip(), "Q47158-P1014-1	Q47158	P1014	300391257	normal")
        self.assertEqual(lines[len(lines) - 1].replace('\n', ''),
                         "Q47158-wikipedia_sitelink-9	Q47158	wikipedia_sitelink	http://et.wikipedia.org/wiki/Triias	")
        f.close()

    def test_kgtk_ifnotexists(self):
        Q47158_path = 'data/Q47158_non_edge.tsv'
        cli_entry("kgtk", "ifnotexists", "-i", self.file_path, "--filter-on", Q47158_path, "-o",
                  f'{self.temp_dir}/Q47158.tsv', "--input-keys", "node1", "--filter-keys", "heading", "--mode", "NONE",
                  "--verbose")

        df = pd.read_csv(f'{self.temp_dir}/Q47158.tsv', sep='\t')

        self.assertEqual(len(df), 169)


    def test_kgtk_ifnotexists_preserve_order(self):
        Q47158_path = 'data/Q47158_non_edge.tsv'
        cli_entry("kgtk", "ifnotexists", "-i", self.file_path, "--filter-on", Q47158_path, "-o",
                  f'{self.temp_dir}/Q47158.tsv', "--input-keys", "node1", "--filter-keys", "heading", "--mode", "NONE",
                  "--preserve-order", "--cache-input", "--verbose")

        f = open(f'{self.temp_dir}/Q47158.tsv')
        lines = f.readlines()
        # strip off the \n
        self.assertEqual(lines[1].strip(), "Q1052204-P18-1	Q1052204	P18	Shengjianbao by buncheduptv in San Mateo, CA.jpg	normal")
        self.assertEqual(lines[len(lines) - 1].replace('\n', ''),
                         "Q65695069-label-1	Q65695069	label	Spring Break Forever\t")
        f.close()



"""
Validate property patterns in a KGTK file.

TODO: Need KgtkWriterOptions.
"""

from argparse import Namespace, SUPPRESS
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'aliases': [ 'filter-properties' ],
        'help': 'Validate property patterns in a KGTK file. ',
        'description': 'Validate property patterns in a KGTK file. ' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert clean-data --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderMode, KgtkReaderOptions
    from kgtk.utils.argparsehelpers import optional_bool
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    parser.add_input_file()
    parser.add_input_file(who="The property pattern definitions.", default_stdin=False,
                          options=["--pattern-file"], dest="pattern_file", metavar="PATTERN_FILE")
    parser.add_output_file(optional=True)
    parser.add_output_file(who="The property pattern reject output.", optional=True,
                          options=["--reject-file"], dest="reject_file", metavar="REJECT_FILE")

    parser.add_argument(      "--presorted", dest="grouped_input",
                              help="Indicate that the input has been presorted (or at least pregrouped) on the node1 column. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False, metavar="True|False")

    parser.add_argument(      "--process-node1-groups", dest="reject_node1_groups",
                              help="When True, process all records for a node1 value " +
                              "as a group. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=True, metavar="True|False")

    parser.add_argument(      "--no-complaints", dest="no_complaints",
                              help="When true, do not print complaints (when rejects are expected). (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False, metavar="True|False")

    parser.add_argument(      "--complain-immediately", dest="complain_immediately",
                              help="When true, print complaints immediately (for debugging). (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False, metavar="True|False")

    parser.add_argument(      "--add-isa-column", dest="add_isa_column",
                              help="When true, add an ISA column to the output and reject files. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False, metavar="True|False")

    parser.add_argument(      "--isa-column-name", dest="isa_column_name", default="isa;node2",
                              help="The name for the ISA column. (default %(default)s)")

    parser.add_argument(      "--autovalidate", dest="autovalidate",
                              help="When true, validate node1 and node2 values before testing them. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=True, metavar="True|False")

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, validate_by_default=True, expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)


def run(input_file: KGTKFiles,
        pattern_file: KGTKFiles,
        output_file: KGTKFiles,
        reject_file: KGTKFiles,
        grouped_input: bool = False,
        reject_node1_groups: bool = False,
        no_complaints: bool = False,
        complain_immediately: bool = False,
        add_isa_column: bool = False,
        isa_column_name: str = "isa;node2",
        autovalidate: bool = True,
        errors_to_stdout: bool = False,
        errors_to_stderr: bool = False,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,
        **kwargs # Whatever KgtkReaderOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from pathlib import Path
    import sys
    
    from kgtk.exceptions import KGTKException
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderMode, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.value.propertypatternvalidator import PropertyPatterns, PropertyPatternValidator
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    input_kgtk_file: Path = KGTKArgumentParser.get_input_file(input_file)
    pattern_kgtk_file: Path = KGTKArgumentParser.get_input_file(pattern_file, default_stdin=False)
    output_kgtk_file: typing.Optional[Path] = KGTKArgumentParser.get_optional_output_file(output_file)
    reject_kgtk_file: typing.Optional[Path] = KGTKArgumentParser.get_optional_output_file(reject_file)

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("--input-file=%s" % str(input_kgtk_file), file=error_file)
        print("--pattern-file=%s" % str(pattern_kgtk_file), file=error_file)
        if output_kgtk_file is not None:
            print("--output-file=%s" % str(output_kgtk_file), file=error_file)
        if reject_kgtk_file is not None:
            print("--reject-file=%s" % str(reject_kgtk_file), file=error_file)
        print("--presorted=%s" % str(grouped_input))
        print("--reject-node1-groups=%s" % str(reject_node1_groups))
        print("--complain-immediately=%s" % str(complain_immediately))
        print("--add-isa-column=%s" % str(add_isa_column))
        print("--isa-column-name=%s" % str(isa_column_name))
        print("--autovalidate=%s" % str(autovalidate))
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    if verbose:
        print("Reading data from '%s'" % str(input_kgtk_file), file=error_file, flush=True)
        print("Reading patterns from '%s'" % str(pattern_kgtk_file), file=error_file, flush=True)
        if output_kgtk_file is not None:
            print("Writing good data to '%s'" % str(output_kgtk_file), file=error_file, flush=True)
        if reject_kgtk_file is not None:
            print("Writing rejected data to '%s'" % str(reject_kgtk_file), file=error_file, flush=True)
                
    try:
        pkr: KgtkReader = KgtkReader.open(pattern_kgtk_file,
                                          error_file=error_file,
                                          mode=KgtkReaderMode.EDGE,
                                          options=reader_options,
                                          value_options=value_options,
                                          verbose=verbose,
                                          very_verbose=very_verbose)
        
        pps: PropertyPatterns = PropertyPatterns.load(pkr,
                                                      value_options,
                                                      error_file=error_file,
                                                      verbose=verbose,
                                                      very_verbose=very_verbose)
        
        kr: KgtkReader = KgtkReader.open(input_kgtk_file,
                                         error_file=error_file,
                                         options=reader_options,
                                         value_options=value_options,
                                         verbose=verbose,
                                         very_verbose=very_verbose)

        output_column_names: typing.List[str] = [ ]
        isa_column_idx: int = -1
        if output_kgtk_file is not None:
            output_column_names = kr.column_names.copy()
            if add_isa_column:
                if isa_column_name in output_column_names:
                    isa_column_idx = output_column_names.index(isa_column_name)
                else:
                    isa_column_idx = len(output_column_names)
                    output_column_names.append(isa_column_name)

        ppv: PropertyPatternValidator = PropertyPatternValidator.new(pps,
                                                                     kr,
                                                                     grouped_input=grouped_input,
                                                                     reject_node1_groups=reject_node1_groups,
                                                                     no_complaints=no_complaints,
                                                                     complain_immediately=complain_immediately,
                                                                     isa_column_idx=isa_column_idx,
                                                                     autovalidate=autovalidate,
                                                                     value_options=value_options,
                                                                     error_file=error_file,
                                                                     verbose=verbose,
                                                                     very_verbose=very_verbose)

        kw: typing.Optional[KgtkWriter] = None
        if output_kgtk_file is not None:
            kw = KgtkWriter.open(output_column_names,
                                 output_kgtk_file,
                                 verbose=verbose, very_verbose=very_verbose)
        
        rkw: typing.Optional[KgtkWriter] = None
        if reject_kgtk_file is not None:
            rkw = KgtkWriter.open(output_column_names,
                                  reject_kgtk_file,
                                  verbose=verbose, very_verbose=very_verbose)
        
        ppv.process(kr, kw, rkw)

        if verbose:
            print("Read %d rows, %d valid" % (ppv.input_row_count, ppv.valid_row_count), file=error_file, flush=True)
            if kw is not None:
                print("Wrote %d good rows" % ppv.output_row_count, file=error_file, flush=True)
            if rkw is not None:
                print("Wrote %d rejected rows" % ppv.reject_row_count, file=error_file, flush=True)

        if kw is not None:
            kw.close()
        if rkw is not None:
            rkw.close()

        return 0

    except Exception as e:
        raise KGTKException(e)




import re
import string


white_space_regex = re.compile('\\s+')
punct_trans_table = dict((ord(char), ord(' ')) for char in string.punctuation)


def normalize_whitespace(text):
    """Strip surrounding whitespace and normalize all other whitespace to single spaces.
    """
    text = str(text)
    return re.sub(white_space_regex, ' ', text).strip()


def normalize_punctuation(text):
    """Translate all punctuation chars in `text' to spaces and then whitespace-normalize the result.
    """
    text = str(text)
    text = text.translate(punct_trans_table)
    return normalize_whitespace(text)


def normalize_case(text):
    """Convert `text' to lower case.
    """
    text = str(text)
    return text.lower()


def normalize_text(text):
    """Perform all text normalizations on `text'.
    """
    return normalize_case(normalize_punctuation(text))


def normalize_ontology_type(type):
    """Normalize an ontology type name by substituting all punctuation characters with spaces.
    This will do the right thing for LDC/AIDA ontology types but might have to be generalized.
    """
    return normalize_punctuation(type)
